"""
    Module that defines the `Results` class
"""

from collections.abc import Collection
from typing import List, Tuple, Iterator
from warnings import warn

import numpy as np


class Results(Collection):
    """
        A class that stores simulation results and provides methods to access them

        Parameters
        ----------
        species_names : List[str]
            List of species names
        rxn_names : List[str]
            List of reaction names
        t_list: List[float]
            List of time points for each repetition
        x_list: List[np.ndarray]
            List of system states for each repetition
        status_list: List[int]
            List of return status for each repetition
        algorithm: str
            Algorithm used to run the simulation
        sim_seeds: List[int]
            List of seeds used for the simulation

        Notes
        -----
        The status indicates the status of the simulation at exit. Each
        repetition will have a status associated with it, and these are
        accessible through the ``status_list``.

        1: Succesful completion, terminated when ``max_iter`` iterations reached.

        2: Succesful completion, terminated when ``max_t`` crossed.

        3: Succesful completion, terminated when all species went extinct.

        -1: Failure, order greater than 3 detected.

        -2: Failure, propensity zero without extinction.
    """

    def __init__(
        self,
        species_names: List[str],
        rxn_names: List[str],
        t_list: List[np.ndarray],
        x_list: List[np.ndarray],
        status_list: List[int],
        algorithm: str,
        sim_seeds: List[int],
    ) -> None:
        self.species_names = species_names
        self.rxn_names = rxn_names
        self.x_list = x_list
        self.t_list = t_list
        self.status_list = status_list
        self.algorithm = algorithm
        self.sim_seeds = sim_seeds
        if not self._check_consistency():
            raise ValueError("Inconsistent results passed")

    def _check_consistency(self) -> bool:
        """
            Check consistency of results

            Returns
            -------
            bool
                True if results are consistent
                False otherwise
        """
        if (
            len(self.x_list)
            == len(self.t_list)
            == len(self.status_list)
            == len(self.sim_seeds)
        ):
            pass
        else:
            return False
        for x, t, status in self:
            if x.shape[0] != t.shape[0]:
                return False
            if x.shape[1] != len(self.species_names):
                return False
            if not isinstance(status, int):
                return False
        return True

    def __repr__(self) -> str:
        """
            Return summary of simulation.

            Returns
            -------
            summary: str
                Summary of the simulation with length of simulation, algorithm and seeds used.
        """
        summary = f"<Results species={self.species_names} n_rep={len(self)} "
        summary = summary + f"algorithm={self.algorithm} sim_seeds={self.sim_seeds}>"
        return summary

    def __str__(self) -> str:
        """ Return self.__repr__() """
        return self.__repr__()

    def __iter__(self) -> Iterator[Tuple[np.ndarray, np.ndarray, int]]:
        """ Iterate over each repetition """
        return zip(self.x_list, self.t_list, self.status_list)

    def __len__(self) -> int:
        """
            Return number of repetitions in simulation

            Returns
            -------
            n_rep: int
                Number of repetitions in simulation
        """
        n_rep = len(self.x_list)
        return n_rep

    def __contains__(self, ind: int):
        """ Returns True if ind is one of the repetition numbers """
        if ind < len(self):
            return True
        else:
            return False

    def __getitem__(self, ind: int) -> Tuple[np.ndarray, np.ndarray, int]:
        """
            Return sim. state, time points and status of repetition no. `ind`

            Parameters
            ----------
            ind: int
                Index of the repetition in the simulation

            Returns
            -------
            x_ind: np.ndarray
                Simulation status of repetition no. `ind`
            t_ind: np.ndarray
                Time points of repetition no. `ind`
            status_ind
                Simulation end status of repetition no. `ind`
        """
        if ind in self:
            x_ind = self.x_list[ind]
            t_ind = self.t_list[ind]
            status_ind = self.status_list[ind]
        else:
            raise IndexError(f"{ind} out of bounds")
        return x_ind, t_ind, status_ind

    @property
    def final(self) -> Tuple[np.ndarray, np.ndarray]:
        """
            Returns the final times and states of the system in the simulations

            Returns
            -------
            Tuple[np.ndarray, np.ndarray]
                The final times and states of the sytem
        """
        final_times = np.array([v[1][-1] for v in self])
        final_states = np.array([v[0][-1, :] for v in self])
        return final_times, final_states

    def get_state(self, t: float) -> List[np.ndarray]:
        """
            Returns the states of the system at time point t.

            Parameters
            ----------
            t: float
                Time point at which states are wanted.

            Returns
            -------
            List[np.ndarray]
                The states of the system at `t` for all repetitions.

            Raises
            ------
            UserWarning
                If simulation ends before `t` but system does not reach
                extinction.
        """
        states: List[np.ndarray] = []
        e = np.finfo(float).eps * t
        t = t + e
        for x_array, t_array, s in self:
            ind = np.searchsorted(t_array, t)
            ind = ind - 1 if ind > 0 else ind
            if ind == len(t_array) - 1:
                states.append(x_array[-1, :])
                if s != 3:
                    warn(f"Simulation ended before {t}, returning last state.")
            else:
                x_interp = np.zeros(x_array.shape[1])
                if self.algorithm != "direct":
                    for ind2 in range(x_array.shape[1]):
                        x_interp[ind2] = np.interp(
                            t,
                            [t_array[ind], t_array[ind + 1]],
                            [x_array[ind, ind2], x_array[ind + 1, ind2]],
                        )
                    states.append(x_interp)
                else:
                    states.append(x_array[ind, :])
        return states

    def get_species(self, species_names: List[str]) -> List[np.ndarray]:
        """
            Returns the species concentrations only for the species in species_names

            Parameters
            ---------
            species_names : List[str]
                The names of the species as a list

            Returns
            ------
            List[np.ndarray]
                Simulation output of the selected species.
        """
        x_list_curated = []
        species_inds = [self.species_names.index(s) for s in species_names]
        for rep_ind in range(len(self)):
            x_list_curated.append(self[rep_ind][0][:, species_inds])
        return x_list_curated



"""
    Contains various utility functions
"""

cimport cython
cimport numpy as np
import numpy as np
# import random


Na = 6.023e23  # Avogadro's constant
HIGH = 1e20
TINY = 1e-20


# @cython.returns((int, int))
@cython.boundscheck(False)
@cython.wraparound(False)
cdef roulette_selection(double [:] prop_list, long long [:] Xt):
    """
        Perform roulette selection on the list of propensities.

        Return the index of the selected reaction (``choice``) by performing
        Roulette selection on the given list of reaction propensities.

        Parameters
        ----------
        prop_list: array_like
            A 1D array of the propensities of the reactions.

        Xt: array_like
            A 1D array of the current simulation state.

        Returns
        -------
        choice: int
            Index of the chosen reaction.
        status: int
            Status of the simulation as described in ``direct``.
    """
    cdef:
        int choice= 0, status=0, counter=0, counter_max=prop_list.shape[0], Xt_counter_max = Xt.shape[0]
        double prop0 = 0.0, Xtsum = 0.0, TINY = 1e-20
    cdef (int, int) result
    for counter in range(counter_max):
        prop0 += prop_list[counter]
    if prop0 < TINY:
        for counter in range(Xt_counter_max):
            Xtsum += Xt[counter]
        counter = 0
        if Xtsum < TINY:
            status = 3
            result = -1, status
            return result
        else:
            status = -2
            result = -1, status
            return result

    cdef double prop_sum = prop_list[0]/prop0
    # r1 = random.random()  # Roll the wheel
    r1 = np.random.random()  # Roll the wheel
    for counter in range(1, counter_max + 1):
        if r1 < prop_sum:
            choice = counter - 1
            break
        else:
            prop_sum += prop_list[counter]/prop0
    result = choice, 0
    return result


def get_kstoc(react_stoic, k_det, volume, chem_flag):
    """
        Compute k_stoc from k_det.

        Return a vector of the stochastic rate constants (k_stoc) determined
        from the deterministic rate constants (k_det) [#]_

        Parameters
        ----------
        react_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the reactants.
            Reactions are columns and species are rows.
        k_det: (nr,) ndarray
            A 1D array representing the deterministic rate constants of the
            system.
        volume: float
            The volume of the reactor vessel which is important for second
            and higher order reactions
        chem_flag: bool
            If True, divide by Na (Avogadro's constant) while calculating
            stochastic rate constants. Defaults to ``False``.

        Returns
        -------
        k_stoc: (nr,) ndarray
            A 1D array representing the stochastic rate constants of the
            system.

        References
        ----------
        .. [#] Gillespie, D.T., 1976.
            A general method for numerically simulating the stochastic time evolution
            of coupled chemical reactions. J. Comput. Phys. 22, 403–434.
            doi:10.1016/0021-9991(76)90041-3.
    """
    cdef:
        int nr=react_stoic.shape[1], ns=react_stoic.shape[0], max_order
        double factor = 1.0
    orders = np.zeros((nr,), dtype=np.int64) # Order of rxn = number of reactants
    for ind1 in range(nr):
        for ind2 in range(ns):
            orders[ind1] += react_stoic[ind2, ind1]
    cdef double[:] k_stoc = k_det.copy()
    if chem_flag:
        factor = Na
    for ind in range(nr):
        max_order = max(react_stoic[:, ind])
        # If highest order is 3
        if max_order == 3:
            k_stoc[ind] = k_det[ind] * 6 / pow(factor * volume, 2)
        elif max_order == 2:  # Highest order is 2
            k_stoc[ind] = k_det[ind] * 2 / pow(factor * volume, orders[ind] - 1)
        else:
            k_stoc[ind] = k_det[ind] / pow(factor * volume, orders[ind] - 1)
    return k_stoc

def py_roulette_selection(prop_list, Xt):
    """
        Perform roulette selection on the list of propensities.

        Return the index of the selected reaction (``choice``) by performing
        Roulette selection on the given list of reaction propensities.

        Parameters
        ----------
        prop_list: array_like
            A 1D array of the propensities of the reactions.

        Xt: array_like
            A 1D array of the current simulation state.

        Returns
        -------
        choice: int
            Index of the chosen reaction.
        status: int
            Status of the simulation as described in ``direct``.
    """
    return roulette_selection(prop_list, Xt)



"""
    Tests for the `Results` class
"""

import numpy as np
import pytest

from cayenne.simulation import Simulation
from cayenne.results import Results


@pytest.mark.parametrize("algorithm", ["direct", "tau_leaping", "tau_adaptive"])
@pytest.mark.usefixtures("setup_large")
class TestResults:
    """
        Test the Results class.
    """

    def test_init_good(self, algorithm, setup_large):
        """
            Test if initialization works.
        """
        species_names, rxn_names, V_r, V_p, X0, k = setup_large
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        with pytest.warns(Warning):
            sim.results
        sim.simulate(algorithm=algorithm)
        assert sim.results

    def test_init_bad(self, algorithm, setup_large):
        """
            Test if initialization fails when it is supposed to.
        """
        species_names, rxn_names, V_r, V_p, X0, k = setup_large
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm=algorithm)
        results = sim.results
        t_list = results.t_list
        x_list = results.x_list
        status_list = results.status_list
        algorithm = "direct"
        seed = [0] * len(status_list)
        assert Results(
            species_names, rxn_names, t_list, x_list, status_list, algorithm, seed
        )
        with pytest.raises(ValueError):
            Results(
                species_names,
                rxn_names,
                t_list,
                x_list,
                status_list,
                algorithm,
                seed[:-2],
            )
        with pytest.raises(ValueError):
            Results(
                species_names,
                rxn_names,
                t_list[:-2],
                x_list,
                status_list,
                algorithm,
                seed,
            )
        with pytest.raises(ValueError):
            t_list[0] = t_list[0][:-2]
            Results(
                species_names, rxn_names, t_list, x_list, status_list, algorithm, seed
            )
        with pytest.raises(ValueError):
            status_list[-1] = "fail"
            Results(
                species_names, rxn_names, t_list, x_list, status_list, algorithm, seed
            )

    def test_iter_len(self, algorithm, setup_large):
        """
            Test the ``__len__`` method and if shape of x and t match.
        """
        species_names, rxn_names, V_r, V_p, X0, k = setup_large
        n_rep = 10
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm=algorithm, n_rep=n_rep)
        results = sim.results
        assert len(results) == n_rep
        for x, t, s in results:
            assert x.shape[0] == t.shape[0]
            assert s

    def test_contains_getitem(self, algorithm, setup_large):
        """
            Test the ``__getitem__`` method.
        """
        species_names, rxn_names, V_r, V_p, X0, k = setup_large
        n_rep = 10
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm=algorithm, n_rep=n_rep)
        results = sim.results
        assert 9 in results
        x, t, s = results[9]
        assert x.shape[0] == t.shape[0]
        assert s
        with pytest.raises(IndexError):
            results[100]

    def test_final(self, algorithm, setup_large):
        """
            Test the ``final`` property.
        """
        species_names, rxn_names, V_r, V_p, X0, k = setup_large
        n_rep = 3
        max_t = 1e5
        max_iter = 100
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm=algorithm, max_t=max_t, max_iter=max_iter, n_rep=n_rep)
        results = sim.results
        final_times, final_states = results.final
        assert final_times.shape[0] == final_states.shape[0]
        if algorithm != "tau_leaping" and algorithm != "tau_leaping":
            for i in range(final_states.shape[0]):
                assert (final_states[0, :] == final_states[i, :]).all()

    def test_get_states(self, algorithm, setup_large):
        """
            Test the ``get_state`` method.
        """
        species_names, rxn_names, V_r, V_p, X0, k = setup_large
        n_rep = 3
        max_t = 1e5
        max_iter = 100
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm=algorithm, max_t=max_t, max_iter=max_iter, n_rep=n_rep)
        results = sim.results
        assert np.isclose(results.get_state(0.0), np.array(X0, dtype=np.float)).all()

        x_list = [
            np.array([0, 1, 2, 3]).reshape(-1, 1),
            np.array([0, 2, 3, 4]).reshape(-1, 1),
            np.array([0, 0, 1, 2]).reshape(-1, 1),
        ]
        t_list = [
            np.array([0, 1, 2, 3]),
            np.array([0, 1, 2, 3]),
            np.array([0, 1, 2, 3]),
        ]
        status_list = [1, 1, 1]
        seed = [0, 1, 2]
        res = Results(["A"], ["r1"], t_list, x_list, status_list, algorithm, seed)
        zero_array = np.array([0])
        one_array = np.array([1])
        two_array = np.array([2])
        three_array = np.array([3])
        four_array = np.array([4])
        assert np.isclose(res.get_state(0), [zero_array, zero_array, zero_array]).all()
        assert np.isclose(res.get_state(1), [one_array, two_array, zero_array]).all()
        assert np.isclose(
            res.get_state(1 + np.finfo(float).eps), [one_array, two_array, zero_array]
        ).all()
        assert np.isclose(res.get_state(3), [three_array, four_array, two_array]).all()

    def test_get_states_high_time(self, algorithm, setup_00003):
        """
            Test the ``get_state`` method at longer model times.
        """
        (
            species_names,
            rxn_names,
            V_r,
            V_p,
            X0,
            k,
            _,
            _,
            _,
            max_t,
            max_iter,
            n_rep,
        ) = setup_00003
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm=algorithm, max_t=10000, max_iter=max_iter, n_rep=1)
        print(
            "status", sim.results.status_list[0], sim.results.t_list, sim.results.x_list
        )
        states = sim.results.get_state(t=15000)
        _, x_final = sim.results.final
        assert states[0] == x_final[0]
        sim.simulate(algorithm=algorithm, max_t=2, max_iter=max_iter, n_rep=1)
        with pytest.warns(UserWarning):
            sim.results.get_state(t=1_000_000)



"""
    Experimental implementation of the `tau adaptive algorithm
    <https://doi.org/10.1063/1.2159468>`_. This is an approximate method that
    builds off the tau_leaping method. It self-adapts the value of ``tau``
    over the course of the simulation. For sytems with a small number of
    molecules, it will be similar in speed to the direct method. For systems
    with a large number of molecules, it will be much faster than the direct
    method.
"""

cimport cython
cimport numpy as np
import numpy as np
# import random
from ..utils import get_kstoc, TINY, HIGH
from ..utils cimport roulette_selection
from .direct import direct
from libc.math cimport log


cdef step1(
        double [:] kstoc_view,
        long long [:] xt_view,
        long [:, :] react_stoic_view,
        long [:, :] v_view,
        int nc,
    ):
    """ Determine critical reactions """
    cdef:
        Py_ssize_t ns=react_stoic_view.shape[0], nr=react_stoic_view.shape[1]
        double prop_sum = 0
        long [:] vis
        int visflag = 0
    cdef double [:] prop_view = np.zeros(nr)
    prop_view[...] = kstoc_view
    L = np.ones(nr, dtype=np.int64) * (nc + 1)
    cdef np.int32_t [:] crit = np.zeros(nr, dtype=np.int32)
    cdef np.int32_t [:] not_crit = np.zeros(nr, dtype=np.int32)
    # Calculate the propensities
    for ind1 in range(nr):
        for ind2 in range(ns):
            # prop = kstoc * product of (number raised to order)
            if react_stoic_view[ind2, ind1] == 1:
                prop_view[ind1] *= xt_view[ind2]
            elif react_stoic_view[ind2, ind1] == 2:
                prop_view[ind1] *= xt_view[ind2] * (xt_view[ind2] - 1) / 2
            elif react_stoic_view[ind2, ind1] == 3:
                prop_view[ind1] *= xt_view[ind2] * (xt_view[ind2] - 1) * (xt_view[ind2] - 2) / 6
    for ind1 in range(nr):
        vis = v_view[:, ind1]
        for ind2 in range(ns):
            if vis[ind2] < 0:
                L[ind1] = min(L[ind1], xt_view[ind2] / abs(vis[ind2]))
    # A reaction j is critical if Lj <nc. However criticality is
    # considered only for reactions with propensity greater than
    # 0 (`prop > 0`).
    for ind1 in range(nr):
        if (L[ind1] < nc) and (prop_view[ind1] > 0):
            crit[ind1] = 1
            not_crit[ind1] = 0
        else:
            crit[ind1] = 0
            not_crit[ind1] = 1
    # crit = (L < nc) * (prop > 0)
    return prop_view, crit, not_crit


def _py_step1(kstoc, xt, react_stoic, v, nc):
    prop_view, crit, not_crit = step1(kstoc, xt, react_stoic, v, nc)
    return np.array(prop_view), np.array(crit), np.array(not_crit)


cdef inline step2_get_g(int hor, long long x):
    cdef double g
    if hor > 0:
        g = float(hor)
    elif hor == -2:
        if x is not 1:
            g = 2.0 + 1.0 / (x - 1.0)
        else:
            g = 2.0
    elif hor == -3:
        if x not in [1, 2]:
            g = 3.0 + 1.0 / (x - 1.0) + 2.0 / (x - 2.0)
        else:
            g = 3.0
    elif hor == -32:
        if x is not 1:
            g = 3.0 / 2.0 * (2.0 + 1.0 / (x - 1.0))
        else:
            g = 3.0
    return g


def _py_step2_get_g(hor, x):
    return step2_get_g(hor, x)

cdef step2(
    int [:] not_crit,
    np.int32_t [:] react_species_view,
    long [:, :] v_view,
    long long [:] xt_view,
    np.int32_t [:] hor_view,
    double [:] prop,
    double epsilon,
    ):

    """ 2. Generate candidate taup """
    cdef:
        Py_ssize_t n_react_species = react_species_view.shape[0]
        Py_ssize_t ns=v_view.shape[0], nr=v_view.shape[1]
    cdef double [:] mup_view = np.zeros(n_react_species, dtype=np.float64)
    cdef double [:] sigp_view = np.zeros(n_react_species, dtype=np.float64)
    cdef double [:] tau_num = np.zeros(n_react_species, dtype=np.float64)
    cdef double taup = HIGH
    cdef int crit_flag = 0
    for ind in range(nr):
        if not_crit[ind] != 0:
            crit_flag = 1
            break
    if not crit_flag:
        taup = HIGH
    else:
        # Compute mu from eqn 32a and sig from eqn 32b
        for ind, species_index in enumerate(react_species_view):
            for i in range(nr):
                if not_crit[i]:
                    mup_view[ind] += v_view[species_index, i] * prop[i]
                    sigp_view[ind] += v_view[species_index, i] * prop[i] * v_view[species_index, i]
            if mup_view[ind] == 0:
                mup_view[ind] = TINY
            if sigp_view[ind] == 0:
                sigp_view[ind] = TINY
            g = step2_get_g(hor_view[species_index], xt_view[species_index])
            tau_num[ind] = max(epsilon * xt_view[species_index] / g, 1)
        for ind in range(n_react_species):
            if mup_view[ind] != 0:
                v1 = tau_num[ind] / abs(mup_view[ind])
            else:
                v1 = HIGH
            if sigp_view[ind] != 0:
                v2 = tau_num[ind] * tau_num[ind] / sigp_view[ind]
            else:
                v2 = HIGH
            taup = min(taup, min(v1, v2))
        # print("mup, sigp, tau_num, xt, taup", np.array(mup_view), np.array(sigp_view), np.array(tau_num), np.array(xt_view), taup)
    return taup


def _py_step2(not_crit,
    react_species,
    v,
    xt,
    hor,
    prop,
    epsilon):
    taup = step2(not_crit, react_species, v, xt, hor, prop, epsilon)
    return np.array(taup)


cdef step5(
    double taup,
    double taupp,
    Py_ssize_t nr,
    int [:] not_crit,
    double [:] prop_view,
    long long [:] xt_view,
    ):
    cdef:
        double tau
        double [:] temp_prop = np.zeros(nr)
        long long [:] K = np.zeros(nr, dtype=np.int64)
    if taup < taupp:
        tau = taup
        for ind in range(nr):
            if not_crit[ind]:
                K[ind] = np.random.poisson(prop_view[ind] * tau)
            else:
                K[ind] = 0
        # print("taup, prop_view, K", taup, np.array(prop_view), np.array(K))
    else:
        tau = taupp
        # Identify the only critical reaction to fire
        # Send in xt to match signature of roulette_selection
        for ind in range(nr):
            if not_crit[ind]:
                temp_prop[ind] = 0.0
            else:
                temp_prop[ind] = prop_view[ind]
        j_crit, _ = roulette_selection(temp_prop, xt_view)
        for ind in range(nr):
            if not_crit[ind]:
                K[ind] = np.random.poisson(prop_view[ind] * tau)
            elif ind == j_crit:
                K[ind] = 1
            else:
                K[ind] = 0
    return tau, K


def _py_step5(taup, taupp, nr, not_crit, prop, xt):
    tau, K = step5(taup, taupp, nr, not_crit, prop, xt)
    return tau, np.array(K)

@cython.boundscheck(False)
@cython.wraparound(False)
def tau_adaptive(
    react_stoic: np.ndarray,
    prod_stoic: np.ndarray,
    init_state: np.ndarray,
    k_det: np.ndarray,
    hor: np.ndarray,
    nc: int,
    epsilon: float,
    max_t: float,
    max_iter: int,
    volume: float,
    seed: int,
    chem_flag: bool,
):
    """
        Runs the adaptive tau leaping simulation algorithm [#]_

        Parameters
        ----------
        react_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the reactants.
            Reactions are columns and species are rows.
        prod_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the products.
            Reactions are columns and species are rows.
        init_state: (ns,) ndarray
            A 1D array representing the initial state of the system.
        k_det: (nr,) ndarray
            A 1D array representing the deterministic rate constants of the
            system.
        hor
            A 1D array of the highest order reaction in which each species
            appears.
        nc: int
            The criticality threshold. Reactions with that cannot fire more than
            nc times are deemed critical.
        epsilon: float
            The epsilon used in tau-leaping, measure of the bound on relative
            change in propensity.
        max_t: float
            The maximum simulation time to run the simulation for.
        max_iter: int
            The maximum number of iterations to run the simulation for.
        volume: float
            The volume of the reactor vessel which is important for second
            and higher order reactions. Defaults to 1 arbitrary units.
        seed: int
            The seed for the numpy random generator used for the current run
            of the algorithm.
        chem_flag: bool
            If True, divide by Na (Avogadro's constant) while calculating
            stochastic rate constants. Defaults to ``False``.

        Returns
        -------
        t: ndarray
            Numpy array of the times.
        x: ndarray
            Numpy array of the states of the system at times in in ``t``.
        status: int
            Indicates the status of the simulation at exit.

            1 - Succesful completion, terminated when ``max_iter`` iterations reached.

            2 - Succesful completion, terminated when ``max_t`` crossed.

            3 - Succesful completion, terminated when all species went extinct.

            -1 - Failure, order greater than 3 detected.

            -2 - Failure, propensity zero without extinction.

            -3 - Negative species count encountered

        References
        ----------
        .. [#] Cao, Y., Gillespie, D.T., Petzold, L.R., 2006.
            Efficient step size selection for the tau-leaping simulation
            method. J. Chem. Phys. 124, 044109. doi:10.1063/1.2159468
    """
    cdef:
        int ite = 1
        long long xtsum = 0
        bint skipflag = 0, negflag = 0
        Py_ssize_t ns=react_stoic.shape[0], nr=react_stoic.shape[1]
        double prop_sum = 0, prop_crit_sum = 0, taup, taupp
    v = prod_stoic - react_stoic  # ns x nr
    x = np.zeros((max_iter, ns), dtype=np.int64)
    t = np.zeros((max_iter))
    x[0, :] = init_state.copy()
    # random.seed(seed)  # Set the seed
    np.random.seed(seed)
    # Determine kstoc from kdet and the highest order or reactions
    prop = get_kstoc(react_stoic, k_det, volume, chem_flag) # Vector of propensities
    kstoc = prop.copy()  # Stochastic rate constants
    vis = np.zeros(nr, dtype=np.int64)
    react_species = np.where(np.sum(react_stoic, axis=1) > 0)[0]
    react_species = react_species.astype(dtype=np.int32)

    cdef double [:] kstoc_view = kstoc
    cdef double [:] prop_view = prop
    cdef long [:, :] v_view = v
    cdef long [:, :] react_stoic_view = react_stoic
    cdef long long [:, :] x_view = x
    cdef double [:] t_view = t
    cdef np.int32_t [:] react_species_view = react_species
    cdef np.int32_t [:] hor_view = hor.astype(np.int32)
    cdef long long [:] vdotK = np.zeros((ns,), dtype=np.int64)
    cdef long long [:] x_new = np.zeros((ns,), dtype=np.int64)


    for ind in range(nr):
        prop_sum += prop_view[ind]
    if prop_sum < TINY:
        for ind in range(ns):
            xtsum += x_view[0, ind]
        if xtsum > TINY:
            status = -2
            return t[:ite], x[:ite, :], status


    while ite < max_iter:
        if not skipflag:

            # Step 1
            # print("starting step 1", ite, t[ite-1])
            prop_view, crit, not_crit = step1(
                kstoc_view,
                x_view[ite - 1, :],
                react_stoic_view,
                v_view,
                nc
            )
            prop_sum = 0
            for ind in range(nr):
                prop_sum += prop_view[ind]
            if prop_sum < TINY:
                status = 3
                return t[:ite], x[:ite, :], status

            # Step 2
            # print("starting step 2", ite, t[ite-1])
            taup = step2(
                not_crit,
                react_species_view,
                v_view,
                x_view[ite-1, :],
                hor_view,
                prop_view,
                epsilon,
            )
            # print("ite, taup, x, 10/propsum, time is: ", ite, taup, np.array(x_view[ite-1, :]), 10.0/prop_sum, t[ite-1])

        # Step 3: For small taup, do SSA
        skipflag = 0
        if taup < 10.0 / prop_sum:
            t_ssa, x_ssa, status = direct(
                react_stoic,
                prod_stoic,
                x[ite - 1, :],
                k_det,
                max_t=max_t - t[ite - 1],
                max_iter=min(101, max_iter - ite),
                volume=volume,
                seed=np.random.randint(low=0, high=1e7),
                chem_flag=chem_flag,
            )
            # t_ssa first element is 0. x_ssa first element is x[ite - 1, :].
            # Both should be dropped while logging the results.
            len_simulation = len(t_ssa) - 1  # Since t_ssa[0] is 0
            for ind1 in range(len_simulation):
                t_view[ite+ind1] = t_ssa[1+ind1] + t[ite-1]
                for ind2 in range(ns):
                    x_view[ite+ind1, ind2] = x_ssa[1+ind1, ind2]
            ite += len_simulation
            if status == 3 or status == 2:
                return t[:ite], x[:ite, :], status
            continue

        # Step 4. Generate second candidate taupp
        # ----------------------------------
        # print("starting step 4", ite, t[ite-1])
        prop_crit_sum = 0.0
        for ind in range(nr):
            if crit[ind]:
                prop_crit_sum += prop[crit[ind]]
        if prop_crit_sum == 0:
            taupp = HIGH
        else:
            # taupp = 1 / prop_crit_sum * log(1 / random.rand())
            taupp = 1 / prop_crit_sum * log(1 / np.random.random())
            # taupp = np.random.exponential(1/prop_crit_sum)

        # Step 5. Leap
        # ------------
        tau, K = step5(taup, taupp, nr, not_crit, prop_view, x_view[ite-1, :])
        # print("taup, taupp, tau", taup, taupp, tau)

        # 6. Handle negatives, update and exit conditions
        # -----------------------------------------------
        # v = ns, nr
        # K = nr
        for ind1 in range(ns):
                vdotK[ind1] = 0
        negflag = 0
        for ind1 in range(ns):
            for ind2 in range(nr):
                vdotK[ind1] += v[ind1, ind2] * K[ind2]
            x_new[ind1] = x[ite - 1, ind1] + vdotK[ind1]
            if x_new[ind1] < 0:
                negflag = 1
        # print("vdotk, k, negflag", np.array(vdotK), np.array(K), negflag)
        if negflag:
            taup = taup / 2
            skip_flag = True
        else:
            # Update states if nothing is negative
            for ind1 in range(ns):
                x_view[ite, ind1] = x_view[ite - 1, ind1] + vdotK[ind1]
            t_view[ite] = t_view[ite - 1] + tau
            ite = ite + 1

        # Exit conditions
        if t[ite - 1] > max_t:
            status = 2
            return t[:ite], x[:ite, :], status

    status = 1
    return t[:ite], x[:ite, :], status



"""Tests for the cython implementation of direct_naive."""

import pytest
import numpy as np
from cayenne.simulation import Simulation
from cayenne.utils import get_kstoc
from cayenne.utils import py_roulette_selection as roulette_selection


@pytest.mark.usefixtures("setup_basic", "setup_large")
class TestCython:
    def test_sim(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm="direct", max_t=1e5, max_iter=int(1e8), chem_flag=False)

    def test_get_kstoc(self, setup_basic):

        # A -> B
        # B -> C
        _, _, vr, _, _, k = setup_basic
        kstoc = np.array(get_kstoc(vr, k, 1.0, False))
        assert (kstoc == k).all()
        kstoc = np.array(get_kstoc(vr, k, 2.0, False))
        assert (kstoc == k).all()
        kstoc = np.array(get_kstoc(vr, k, 2.0, True))
        assert (kstoc == k).all()

        # A + B -> C
        # B + C -> C
        vr = np.array([[1, 1, 0], [1, 1, 0]]).transpose()
        kstoc = np.array(get_kstoc(vr, k, 1.0, False))
        V = 3.0
        print(kstoc)
        assert (kstoc == [k[0], k[1]]).all()
        kstoc = np.array(get_kstoc(vr, k, V, False))
        assert (kstoc == [k[0] / V, k[1] / V]).all()

        # A + B -> C
        # A -> C
        vr = np.array([[1, 1, 0], [1, 0, 0]]).transpose()
        kstoc = np.array(get_kstoc(vr, k, 1.0, False))
        assert (kstoc == [k[0], k[1]]).all()
        kstoc = np.array(get_kstoc(vr, k, V, False))
        assert (kstoc == [k[0] / V, k[1]]).all()

        # A -> C
        # A + B -> C
        vr = np.array([[1, 0, 0], [1, 1, 0]]).transpose()
        kstoc = np.array(get_kstoc(vr, k, 1.0, False))
        assert (kstoc == [k[0], k[1]]).all()
        kstoc = np.array(get_kstoc(vr, k, V, False))
        assert (kstoc == [k[0], k[1] / V]).all()

        # A + B -> C
        vr = np.array([[1, 1, 0]]).transpose()
        k = np.array([1.0])
        kstoc = np.array(get_kstoc(vr, k, 1.0, False))
        assert kstoc == k
        kstoc = np.array(get_kstoc(vr, k, V, False))
        assert kstoc == k / V

        # A + B + C -> D
        # A -> B
        vr = np.array([[1, 1, 1, 0], [1, 0, 0, 0]]).transpose()
        k = np.array([1.0, 1.0])
        kstoc = np.array(get_kstoc(vr, k, 1.0, False))
        assert (kstoc == [k[0], k[1]]).all()
        kstoc = np.array(get_kstoc(vr, k, V, False))
        assert (kstoc == [k[0] / V ** 2, k[1]]).all()

        # A + B -> C
        # A + B + C -> D
        vr = np.array([[1, 1, 0, 0], [1, 1, 1, 0]]).transpose()
        k = np.array([1.0, 1.0])
        kstoc = np.array(get_kstoc(vr, k, 1.0, False))
        assert (kstoc == [k[0], k[1]]).all()
        kstoc = np.array(get_kstoc(vr, k, V, False))
        assert (kstoc == [k[0] / V, k[1] / V ** 2]).all()


class TestRoulette:
    def test_roulettecy(self):
        choice, status = roulette_selection(
            np.array([5, 0, 0], dtype=np.double), np.array([1, 2], dtype=np.int64)
        )
        assert choice == 0
        assert status == 0
        choice, status = roulette_selection(
            np.array([0, 5, 0], dtype=np.double), np.array([1, 2], dtype=np.int64)
        )
        assert choice == 1
        assert status == 0
        choice, status = roulette_selection(
            np.array([0, 0, 5], dtype=np.double), np.array([1, 2], dtype=np.int64)
        )
        assert choice == 2
        assert status == 0
        choice, status = roulette_selection(
            np.array([0, 0, 0], dtype=np.double), np.array([0, 0], dtype=np.int64)
        )
        assert choice == -1
        assert status == 3
        choice, status = roulette_selection(
            np.array([0, 0, 0], dtype=np.double), np.array([1, 2], dtype=np.int64)
        )
        assert choice == -1
        assert status == -2

    def test_100(self):
        prop_list = np.array([1.0, 0.0, 0.0])
        xt = np.array([1, 2], dtype=np.int64)
        choice, status = roulette_selection(prop_list, xt)
        assert status == 0
        assert choice == 0

    def test_010(self):
        prop_list = np.array([0.0, 1.0, 0.0])
        xt = np.array([1, 2], dtype=np.int64)
        choice, status = roulette_selection(prop_list, xt)
        assert status == 0
        assert choice == 1

    def test_001(self):
        prop_list = np.array([0.0, 0.0, 1.0])
        xt = np.array([1, 2], dtype=np.int64)
        choice, status = roulette_selection(prop_list, xt)
        assert status == 0
        assert choice == 2

    def test_stat3(self):
        prop_list = np.array([0.0, 0.0, 0.0])
        xt = np.array([0, 0], dtype=np.int64)
        _, status = roulette_selection(prop_list, xt)
        assert status == 3

    def test_statm2(self):
        prop_list = np.array([0.0, 0.0, 0.0])
        xt = np.array([1, 0], dtype=np.int64)
        _, status = roulette_selection(prop_list, xt)
        assert status == -2



"""
    Common configuration for all the tests
"""

import csv
import pathlib

import numpy as np
import pytest

from cayenne.utils import Na


CURR_DIR = pathlib.Path(__file__).parent


@pytest.fixture
def setup_basic():
    """
        Setup the basic system.

        Notes
        -----
        A --> B; k = 1.0
        B --> C; k = 1.0

        A0 = 100, B0 = C0 = 0
    """
    V_r = np.array([[1, 0], [0, 1], [0, 0]])
    V_p = np.array([[0, 0], [1, 0], [0, 1]])
    X0 = np.array([100, 0, 0], dtype=np.int64)
    k = np.array([1.0, 1.0])
    species_names = ["A", "B", "C"]
    rxn_names = ["r1", "r2"]
    return species_names, rxn_names, V_r, V_p, X0, k


@pytest.fixture
def setup_large():
    """
        Setup the large system.

        Notes
        -----
        A --> B; k = 1.0
        B --> C; k = 1.0
        C --> D; k = 1.0
        D --> E; k = 1.0
        E --> _; k = 1.0

        A0 = 10, B0 = C0 = D0 = E0 = 0
    """
    V_r = np.array(
        [
            [1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1],
        ]
    )
    V_p = np.array(
        [
            [0, 0, 0, 0, 0],
            [1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0],
        ]
    )
    X0 = np.array([10, 0, 0, 0, 0], dtype=np.int64)
    k = np.array([1.0, 1.0, 1.0, 1.0, 1.0])
    species_names = ["A", "B", "C", "D", "E"]
    rxn_names = ["r1", "r2", "r3", "r4", "r5"]
    return species_names, rxn_names, V_r, V_p, X0, k


@pytest.fixture
def setup_system():
    """
        Setup minimal system.

        Contains only ``k_det`` and ``volume``. Used by TestKstoc class as
        filler.
    """
    k_det = np.array([3.0])
    volume = 7.0
    return k_det, volume


@pytest.fixture
def setup_bifurcation():
    """
        Setup bifurcation system.

        Notes
        -----
        A --> B; k = 1.0
        B + D --> B + B; k = 0.01*Na
        A + B --> C; k = 1.0

        A0 = 1, B0 = C0 = 0, D0 = 10
    """
    V_r = np.array([[1, 0, 1], [0, 1, 0], [0, 0, 0], [0, 1, 0]])
    V_p = np.array([[0, 0, 0], [1, 2, 0], [0, 0, 1], [0, 0, 0]])
    k = np.array([1.0, 0.01 * Na, 1.0])
    X0 = np.array([1, 0, 0, 10], dtype=np.int64)
    species_names = ["A", "B", "C", "D"]
    rxn_names = ["r1", "r2", "r3"]
    return species_names, rxn_names, V_r, V_p, X0, k


@pytest.fixture
def setup_long():
    """
        Setup a system that runs for a long time.

        Notes
        -----
        A --> B; k = 1.0
        B --> C; k = 1.0

        A0 = 4e5, B0 = 1000, C0 = 0
    """
    V_r = np.array([[1, 0], [0, 1], [0, 0]])
    V_p = np.array([[0, 0], [1, 0], [0, 1]])
    k = np.array([1.0, 1.0])
    X0 = np.array([int(4e5), 1000, 0], dtype=np.int64)
    species_names = ["A", "B", "C"]
    rxn_names = ["r1", "r2"]
    return species_names, rxn_names, V_r, V_p, X0, k


def read_results(test_id: str):
    """
        Read the simulation results used for accuracy tests.

        Parameters
        ----------
        test_id: str
            The index of the test number to return the results of.

        Returns
        -------
        time_list: List[float]
            Time points at which species amounts are analytically predicted.
        mu_list: List[float]
            Time course of the analytically predicted species amount means.
        std_list: List[float]
            Time course of the analytically predicted species amount standard
            deviations.

        See Also
        --------
        read_results_2sp: Read results for 2 species.

        Notes
        -----
        The accuracy tests are taken from the SBML Test Suite [1]_ .

        References
        ----------
        .. [1] https://github.com/sbmlteam/sbml-test-suite/tree/master/cases/stochastic
    """
    filename = f"{CURR_DIR}/data/results_{test_id}.csv"
    time_list = []
    mu_list = []
    std_list = []
    with open(filename) as fid:
        csv_reader = csv.reader(fid, delimiter=",")
        next(csv_reader)
        for time, mu, std in csv_reader:
            time_list.append(float(time))
            mu_list.append(float(mu))
            std_list.append(float(std))
    return time_list, mu_list, std_list


def read_results_2sp(test_id: str):
    """
        Read the simulation results used for accuracy tests when 2 species
        are tracked.

        Parameters
        ----------
        test_id: str
            The index of the test number to return the results of.

        Returns
        -------
        time_list: List[float]
            Time points at which species amounts are analytically predicted.
        mu_list: List[float]
            Time course of the analytically predicted species amount means.
        std_list: List[float]
            Time course of the analytically predicted species amount standard
            deviations.

        See Also
        --------
        read_results: Read results for one species.

        Notes
        -----
        The accuracy tests are taken from the SBML Test Suite [1]_ .

        References
        ----------
        .. [1] https://github.com/sbmlteam/sbml-test-suite/tree/master/cases/stochastic
    """
    filename = f"{CURR_DIR}/data/results_{test_id}.csv"
    time_list = []
    mu_list = []
    std_list = []
    with open(filename) as fid:
        csv_reader = csv.reader(fid, delimiter=",")
        next(csv_reader)
        for time, mu1, mu2, std1, std2 in csv_reader:
            time_list.append(float(time))
            mu_list.append(np.array([float(mu1), float(mu2)]))
            std_list.append(np.array([float(std1), float(std2)]))
    return time_list, mu_list, std_list


@pytest.fixture
def setup_00001():
    """
        Setup the accuracy test 00001.

        Notes
        -----
        A --> A + A; k = 0.1
        A --> _; k = 0.11

        A0 = 100, max_t = 51, max_iter = 1.5e3
    """
    V_r = np.array([[1, 1]])
    V_p = np.array([[2, 0]])
    X0 = np.array([100], dtype=np.int64)
    k = np.array([0.1, 0.11])
    max_t = 51
    max_iter = int(1.5e3)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00001")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00003():
    """
        Setup the accuracy test 00003.

        Notes
        -----
        A --> A + A; k = 1.0
        A --> _; k = 1.1

        A0 = 100, max_t = 51, max_iter = 1e5
    """
    species_names = ["A"]
    rxn_names = ["r1", "r2"]
    V_r = np.array([[1, 1]])
    V_p = np.array([[2, 0]])
    X0 = np.array([100], dtype=np.int64)
    k = np.array([1.0, 1.1])
    max_t = 51
    max_iter = int(1e5)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00003")
    return (
        species_names,
        rxn_names,
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00004():
    """
        Setup the accuracy test 00004.

        Notes
        -----
        A --> A + A; k = 0.1
        A --> _; k = 0.11

        A0 = 10, max_t = 51, max_iter = 1.5e3
    """
    V_r = np.array([[1, 1]])
    V_p = np.array([[2, 0]])
    X0 = np.array([10], dtype=np.int64)
    k = np.array([0.1, 0.11])
    max_t = 51
    max_iter = int(1.5e3)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00004")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00005():
    """
        Setup the accuracy test 00005.

        Notes
        -----
        A --> A + A; k = 0.1
        A --> _; k = 0.11

        A0 = 10_000, max_t = 51, max_iter = 5e5
    """
    V_r = np.array([[1, 1]])
    V_p = np.array([[2, 0]])
    X0 = np.array([10000], dtype=np.int64)
    k = np.array([0.1, 0.11])
    max_t = 51
    max_iter = int(5e5)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00005")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00011():
    """
        Setup the accuracy test 00011.

        Notes
        -----
        A --> A + A; k = 0.1/2
        A --> _; k = 0.11/2

        A0 = 100, max_t = 51, max_iter = 1.5e3
    """
    V_r = np.array([[1, 1]])
    V_p = np.array([[2, 0]])
    X0 = np.array([100], dtype=np.int64)
    # divide k by 2 because rate expression given in units of concentration
    # in the model file
    k = np.array([0.1, 0.11]) / 2
    max_t = 51
    max_iter = int(1.5e3)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00011")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00020():
    """
        Setup the accuracy test 00020.

        Notes
        -----
        _ --> A; k = 1
        A --> _; k = 0.1

        A0 = 0, max_t = 52, max_iter = 1.5e3
    """
    V_r = np.array([[0, 1]])
    V_p = np.array([[1, 0]])
    X0 = np.array([0], dtype=np.int64)
    k = np.array([1.0, 0.1])
    max_t = 52
    # we did 52 because direct would stop earlier than 50, but we want at least 50 crossed.
    max_iter = int(1.5e3)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00020")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00021():
    """
        Setup the accuracy test 00021.

        Notes
        -----
        _ --> A; k = 10.0
        A --> _; k = 0.1

        A0 = 0, max_t = 51, max_iter = 1.5e3
    """
    V_r = np.array([[0, 1]])
    V_p = np.array([[1, 0]])
    X0 = np.array([0], dtype=np.int64)
    k = np.array([10.0, 0.1])
    max_t = 51
    max_iter = int(1.5e3)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00021")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00022():
    """
        Setup the accuracy test 00022.

        Notes
        -----
        _ --> A; k = 5.0
        A --> _; k = 0.1

        A0 = 0, max_t = 51, max_iter = 1.5e3
    """
    V_r = np.array([[0, 1]])
    V_p = np.array([[1, 0]])
    X0 = np.array([0], dtype=np.int64)
    k = np.array([5.0, 0.1])
    max_t = 51
    max_iter = int(1.5e3)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00022")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00023():
    """
        Setup the accuracy test 00023.

        Notes
        -----
        _ --> A; k = 1000
        A --> _; k = 0.1

        A0 = 0, max_t = 51, max_iter = 1.5e5
    """
    V_r = np.array([[0, 1]])
    V_p = np.array([[1, 0]])
    X0 = np.array([0], dtype=np.int64)
    k = np.array([1000.0, 0.1])
    max_t = 51
    max_iter = int(1.5e5)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00023")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00030():
    """
        Setup the accuracy test 00030.

        Notes
        -----
        A + A --> A2; k = 0.001
        A2 --> A + A; k = 0.01

        A0 = 100, max_t = 55, max_iter = 1.5e5

        In the model description, they just say k1 = 0.001 without specifying
        deterministic or stochastic. They end up using k1_stoc = 0.001. To have
        k1_stoc = 0.001, we should set k1_det = 0.001/2.
    """
    V_r = np.array([[2, 0], [0, 1]])
    V_p = np.array([[0, 2], [1, 0]])
    X0 = np.array([100, 0], dtype=np.int64)
    k = np.array([0.001 / 2, 0.01])
    max_t = 55
    max_iter = int(1.5e5)
    n_rep = 10
    time_list, mu_list, std_list = read_results_2sp("00030")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00031():
    """
        Setup the accuracy test 00031.

        Notes
        -----
        A + A --> A2; k = 0.0002
        A2 --> A + A; k = 0.004

        A0 = 1000, max_t = 52, max_iter = 1.5e5

        In the model description, they just say k1 = 0.0002 without specifying
        deterministic or stochastic. They end up using k1_stoc = 0.0002. To
        have k1_stoc = 0.0002, we should set k1_det = 0.0002/2.
    """
    V_r = np.array([[2, 0], [0, 1]])
    V_p = np.array([[0, 2], [1, 0]])
    X0 = np.array([1000, 0], dtype=np.int64)
    k = np.array([0.0002 / 2, 0.004])
    max_t = 52
    max_iter = int(1.5e5)
    n_rep = 10
    time_list, mu_list, std_list = read_results_2sp("00031")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00037():
    """
        Setup the accuracy test 00037.

        Notes
        -----
        _ --> 5A; k = 1.0
        A --> _; k = 0.2

        A0 = 0, max_t = 53, max_iter = 1.5e3
    """
    V_r = np.array([[0, 1]])
    V_p = np.array([[5, 0]])
    X0 = np.array([0], dtype=np.int64)
    k = np.array([1.0, 0.2])
    max_t = 53
    max_iter = int(1.5e3)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00037")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00038():
    """
        Setup the accuracy test 00038.

        Notes
        -----
        _ --> 10A; k = 1.0
        A --> _; k = 0.4

        A0 = 0, max_t = 53, max_iter = 1.5e3

        In the model description, they just say k2 = 0.2 without specifying
        deterministic or stochastic. This is a first order reaction, so
        technically k2 = k2_stoc. They instead use kstoc = 0.4 in the model
        file, so this value is used.
    """
    V_r = np.array([[0, 1]])
    V_p = np.array([[10, 0]])
    X0 = np.array([0], dtype=np.int64)
    k = np.array([1.0, 0.4])
    max_t = 53
    max_iter = int(1.5e3)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00038")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00039():
    """
        Setup the accuracy test 00038.

        Notes
        -----
        _ --> 100A; k = 1.0
        A --> _; k = 4.0

        A0 = 0, max_t = 53, max_iter = 1.5e5

        In the model description, they just say k2 = 0.2 without specifying
        deterministic or stochastic. This is a first order reaction, so
        technically k2 = k2_stoc. They instead use kstoc = 4.0 in the model
        file, so this value is used.
    """
    V_r = np.array([[0, 1]])
    V_p = np.array([[100, 0]])
    X0 = np.array([0], dtype=np.int64)
    k = np.array([1.0, 4.0])
    max_t = 53
    max_iter = int(1.5e5)
    n_rep = 10
    time_list, mu_list, std_list = read_results("00039")
    return (
        V_r,
        V_p,
        X0,
        k,
        time_list,
        np.array(mu_list),
        np.array(std_list),
        max_t,
        max_iter,
        n_rep,
    )


@pytest.fixture
def setup_00001_correct():
    MODEL_00001 = """
        const compartment comp1;
        comp1 = 7; # volume

        r1: S1 => S1 + S1; k1;
        r2: S1 => ; k2;

        k1 = 0.1;
        k2 = 0.11;
        chem_flag = false;

        S1 = 100;
    """
    return MODEL_00001


@pytest.fixture
def setup_00001_nochemflag():
    MODEL_00001 = """
        const compartment comp1;
        comp1 = 7; # volume

        r1: S1 => S1 + S1; k1;
        r2: S1 => ; k2;

        k1 = 0.1;
        k2 = 0.11;

        S1 = 100;
    """
    return MODEL_00001


@pytest.fixture
def setup_00001_norate():
    MODEL_00001 = """
        const compartment comp1;
        comp1 = 7; # volume

        r1: S1 => S1 + S1; k1;
        r2: S1 => ;

        k1 = 0.1;
        chem_flag = false;

        S1 = 100;
    """
    return MODEL_00001


@pytest.fixture
def setup_00001_noratevalue():
    MODEL_00001 = """
        const compartment comp1;
        comp1 = 7; # volume

        r1: S1 => S1 + S1; k1;
        r2: S1 => ; k2;

        k1 = 0.1;
        chem_flag = false;

        S1 = 100;
    """
    return MODEL_00001


@pytest.fixture
def setup_00001_nospeciesvalue():
    MODEL_00001 = """
        const compartment comp1;
        comp1 = 7; # volume

        r1: S1 => S1 + S1; k1;
        r2: S1 => ; k2;

        k1 = 0.1;
        k2 = 0.11;
        chem_flag = false;

    """
    return MODEL_00001


@pytest.fixture
def setup_00001_rateequation():
    MODEL_00001 = """
        const compartment comp1;
        comp1 = 7; # volume

        r1: S1 => S1 + S1; k1*S1;
        r2: S1 => ; k2*S1;

        k1 = 0.1;
        k2 = 0.11;
        chem_flag = false;

        S1 = 100;
    """
    return MODEL_00001


@pytest.fixture
def setup_00001_incompletespeciesvalue():
    MODEL_00001 = """
        const compartment comp1;
        comp1 = 7; # volume

        r1: S1 => S1 + S1; k1;
        r2: S1 => S2; k2;

        k1 = 0.1;
        k2 = 0.11;
        chem_flag = false;

        S1 = 100;

    """
    return MODEL_00001


@pytest.fixture
def setup_00001_nocompartment():
    MODEL_00001 = """
        r1: S1 => S1 + S1; k1;
        r2: S1 => ; k2;

        k1 = 0.1;
        k2 = 0.11;
        chem_flag = false;

        S1 = 100;
    """
    return MODEL_00001



"""
    Accuracy tests of the algorithms
"""
import numpy as np
import pytest
from cayenne.simulation import Simulation
import matplotlib.pyplot as plt
from cayenne.algorithms.tau_adaptive import tau_adaptive as tac


# def calculate_zy(sim: Simulation, time_list: list, mu_list: list, std_list: list):
#     z_list = []
#     y_list = []
#     n_rep = len(sim.results.t_list)
#     for ind1, t in enumerate(time_list[1:]):
#         results = sim.results.get_state(t)
#         mu_obs = np.mean(results)
#         std_obs = np.std(results)
#         z_list.append(
#             np.sqrt(n_rep) * (mu_obs - mu_list[ind1 + 1]) / std_list[ind1 + 1]
#         )
#         y_list.append(
#             np.sqrt(n_rep / 2) * ((std_obs ** 2) / (std_list[ind1 + 1] ** 2) - 1)
#         )
#     return np.array(z_list), np.array(y_list)


# def calculate_zy_2sp(sim: Simulation, time_list: list, mu_list: list, std_list: list):
#     z_list = []
#     y_list = []
#     n_rep = len(sim.results.t_list)
#     for ind1, t in enumerate(time_list[1:]):
#         results = sim.results.get_state(t)
#         mu_obs = np.mean(results, axis=0)
#         std_obs = np.std(results, axis=0)
#         z_list.append(
#             np.sqrt(n_rep) * (mu_obs - mu_list[ind1 + 1]) / std_list[ind1 + 1]
#         )
#         y_list.append(
#             np.sqrt(n_rep / 2) * ((std_obs ** 2) / (std_list[ind1 + 1] ** 2) - 1)
#         )
#     return np.array(z_list), np.array(y_list)


# def test_1x(setup_00003):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00003
#     sim1 = Simulation(V_r, V_p, X0, k)
#     hor = sim1.HOR
#     nc = 10
#     epsilon = 0.03
#     np.random.seed(0)
#     sim_seeds = np.random.randint(low=0, high=1e7, size=n_rep)
#     print(sim_seeds[4])
#     t, x, status = tac(
#         react_stoic=V_r,
#         prod_stoic=V_p,
#         init_state=X0,
#         k_det=k,
#         hor=hor,
#         nc=nc,
#         epsilon=epsilon,
#         max_t=max_t,
#         max_iter=max_iter,
#         volume=1.0,
#         seed=sim_seeds[4],
#         chem_flag=False,
#     )
#     plt.plot(t, x)
#     plt.show()
#     # print(x)


# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00001(setup_00001, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00001
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm, max_t=max_t, max_iter=max_iter, chem_flag=False, n_rep=20
#     )
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert np.all(sim1.results.final[0] > time_list[-1])
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# def plot_zy(Z, Y, time_list):
#     plt.figure()
#     plt.subplot(121)
#     plt.plot(time_list[1:], Z)
#     plt.hlines(3, 0, 50)
#     plt.hlines(-3, 0, 50)
#     plt.ylabel("Observed Z")
#     plt.subplot(122)
#     plt.plot(time_list[1:], Y)
#     plt.hlines(5, 0, 50)
#     plt.hlines(-5, 0, 50)
#     plt.ylabel("Observed Y")


# def plot_sd(sim, time_list, mu_list, std_list, Z, Y):
#     mu_obs_list = []
#     std_obs_list = []

#     for ind1, t in enumerate(time_list[1:]):
#         results = sim.results.get_state(t)
#         mu_obs_list.append(np.mean(results))
#         std_obs_list.append(np.std(results))

#     plt.figure()
#     plt.subplot(121)
#     plt.plot(time_list[1:], mu_obs_list)
#     for i in range(len(Z)):
#         if -3 <= Z[i] <= 3:
#             marker = "green"
#         else:
#             marker = "red"
#         plt.plot(time_list[i + 1], mu_list[i + 1], "o", color=marker)
#     plt.hlines(0, 0, 50)
#     plt.ylabel("Observed mean")
#     plt.subplot(122)
#     plt.plot(time_list[1:], std_obs_list)
#     for i in range(len(Y)):
#         if -5 <= Y[i] <= 5:
#             marker = "green"
#         else:
#             marker = "red"
#         plt.plot(time_list[i + 1], std_list[i + 1], "o", color=marker)
#     plt.ylabel("Observed SD")


# @pytest.mark.filterwarnings("ignore::UserWarning")
# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00003(setup_00003, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00003
#     sim1 = Simulation(V_r, V_p, X0, k)
#     n_rep = 100
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     t_final, x_final = sim1.results.final
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     # plot_zy(Z, Y, time_list)
#     # plot_sd(sim1, time_list, mu_list, std_list, Z, Y)
#     # plt.show()
#     assert (np.array(sim1.results.status_list) != 1).all()
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (Y < 5).all()
#     assert (-5 < Y).all()


# @pytest.mark.filterwarnings("ignore::UserWarning")
# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00004(setup_00004, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00004
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     t_final, x_final = sim1.results.final
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (np.array(sim1.results.status_list) != 1).all()
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.filterwarnings("ignore::UserWarning")
# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00005(setup_00005, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00005
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     t_final, x_final = sim1.results.final
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (np.array(sim1.results.status_list) != 1).all()
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00011(setup_00011, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00011
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00020(setup_00020, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00020
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00021(setup_00021, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00021
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00022(setup_00022, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00022
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00023(setup_00023, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00023
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.filterwarnings("ignore::UserWarning")
# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00030(setup_00030, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00030
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy_2sp(sim1, time_list, mu_list, std_list)
#     assert (np.array(sim1.results.status_list) != 1).all()
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.filterwarnings("ignore::UserWarning")
# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00031(setup_00031, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00031
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy_2sp(sim1, time_list, mu_list, std_list)
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00037(setup_00037, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00037
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00038(setup_00038, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00038
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()


# @pytest.mark.filterwarnings("ignore::UserWarning")
# @pytest.mark.parametrize("algorithm", ["tau_adaptive"])
# def test_00039(setup_00039, algorithm):
#     V_r, V_p, X0, k, time_list, mu_list, std_list, max_t, max_iter, n_rep = setup_00039
#     sim1 = Simulation(V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm=algorithm,
#         max_t=max_t,
#         max_iter=max_iter,
#         chem_flag=False,
#         n_rep=n_rep,
#     )
#     Z, Y = calculate_zy(sim1, time_list, mu_list, std_list)
#     assert (-3 < Z).all()
#     assert (Z < 3).all()
#     assert (-5 < Y).all()
#     assert (Y < 5).all()



# -*- coding: utf-8 -*-

"""Top-level package for Stochastic Simulation Algorithms in Python."""

__author__ = """Dileep Kishore, Srikiran Chandrasekaran"""
__email__ = "k.dileep1994@gmail.com"
__version__ = "1.0.1"

from .simulation import Simulation



"""
    Implementation of Gillespie's Direct method. This is an exact simulation
    algorithm that simulates each reaction step. This makes it slower than
    other methods, but it's a good place to start.
"""

cimport cython
cimport numpy as np
import numpy as np
# import random # faster than using np.random, but reproducibility issues
from ..utils cimport roulette_selection
from ..utils import get_kstoc
from libc.math cimport log


@cython.boundscheck(False)
@cython.wraparound(False)
def direct(
    react_stoic: np.ndarray,
    prod_stoic: np.ndarray,
    init_state: np.ndarray,
    k_det: np.ndarray,
    max_t: float,
    max_iter: int,
    volume: float,
    seed: int,
    chem_flag: bool,
):
    """
        Runs the Direct Stochastic Simulation Algorithm [#]_

        Parameters
        ----------
        react_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the reactants.
            Reactions are columns and species are rows.
        prod_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the products.
            Reactions are columns and species are rows.
        init_state: (ns,) ndarray
            A 1D array representing the initial state of the system.
        k_det: (nr,) ndarray
            A 1D array representing the deterministic rate constants of the
            system.
        max_t: float
            The maximum simulation time to run the simulation for.
        max_iter: int
            The maximum number of iterations to run the simulation for.
        volume: float
            The volume of the reactor vessel which is important for second
            and higher order reactions. Defaults to 1 arbitrary units.
        seed: int
            The seed for the numpy random generator used for the current run
            of the algorithm.
        chem_flag: bool
            If True, divide by Na (Avogadro's constant) while calculating
            stochastic rate constants. Defaults to ``False``.

        Returns
        -------
        t: ndarray
            Numpy array of the times.
        x: ndarray
            Numpy array of the states of the system at times in in ``t``.
        status: int
            Indicates the status of the simulation at exit.

            1 - Succesful completion, terminated when ``max_iter`` iterations reached.

            2 - Succesful completion, terminated when ``max_t`` crossed.

            3 - Succesful completion, terminated when all species went extinct.

            -1 - Failure, order greater than 3 detected.

            -2 - Failure, propensity zero without extinction.

        References
        ----------
        .. [#] Gillespie, D.T., 1976.
            A general method for numerically simulating the stochastic time
            evolution of coupled chemical reactions. J. Comput. Phys. 22, 403–434.
            doi:10.1016/0021-9991(76)90041-3.
    """

    cdef:
        int ite=1, ind=0, ind1=0, ind2=0, continue_flag=0, choice, status
        double t_curr=0, prop_sum=0
        Py_ssize_t ns=react_stoic.shape[0], nr=react_stoic.shape[1]
    v = prod_stoic - react_stoic  # ns x nr
    x = np.zeros((max_iter, ns), dtype=np.int64)
    t = np.zeros((max_iter))
    x[0, :] = init_state.copy()
    xtemp = init_state.copy()  # Temporary X for updating
    status = 0
    np.random.seed(seed)
    # random.seed(seed)  # Set the seed
    # Determine kstoc from kdet and the highest order or reactions
    prop = get_kstoc(react_stoic, k_det, volume, chem_flag)  # Vector of propensities
    kstoc = prop.copy()  # Stochastic rate constants
    cdef double [:] kstoc_view = kstoc
    cdef double [:] prop_view = prop
    cdef long long [:] xtemp_view = xtemp
    cdef long [:, :] v_view = v
    cdef long [:, :] react_stoic_view = react_stoic
    cdef long long [:, :] x_view = x
    while ite < max_iter:
        # Calculate propensities
        for ind1 in range(nr):
            for ind2 in range(ns):
                # prop = kstoc * product of (number raised to order)
                if react_stoic_view[ind2, ind1]:
                    if react_stoic_view[ind2, ind1] == 1:
                        prop_view[ind1] *= x_view[ite - 1, ind2]
                    elif react_stoic_view[ind2, ind1] == 2:
                        prop_view[ind1] *= x_view[ite - 1, ind2] * (x_view[ite - 1, ind2] - 1) / 2
                    elif react_stoic_view[ind2, ind1] == 3:
                        prop_view[ind1] *= x_view[ite - 1, ind2] * (x_view[ite - 1, ind2] - 1) * (x_view[ite - 1, ind2] - 2) / 6
        # Roulette wheel
        roulette_results = roulette_selection(prop_view, x_view[ite-1, :])
        choice = roulette_results[0]
        status = roulette_results[1]
        if status == 0:
            for ind1 in range(ns):
                xtemp_view[ind1] = x_view[ite-1, ind1] + v_view[ind1, choice]
        else:
            return t[:ite], x[:ite, :], status

        # If negative species produced, reject step
        # if np.min(xtemp) < 0:
        #     continue
        for ind in range(ns):
            if xtemp_view[ind] < 0:
                continue_flag = 1
                break
        if continue_flag:
            continue_flag = 0
            continue
        # Update xt and t_curr
        else:
            r2 = np.random.random()
            prop_sum = 0
            for ind in range(nr):
                prop_sum += prop_view[ind]
            # t_curr += 1 / prop_sum * log(1 / r2)
            # t_curr += -1 / prop_sum * log(r2)
            t_curr += np.random.exponential(1/prop_sum)
            if t_curr > max_t:
                status = 2
                return t[:ite], x[:ite, :], status
        for ind in range(nr):
            prop_view[ind] = kstoc_view[ind]
        for ind in range(ns):
            x_view[ite, ind] = xtemp_view[ind]
        t[ite] = t_curr
        ite += 1
    status = 1
    return t[:ite], x[:ite, :], status



import numpy as np
import pytest

from cayenne.model_io import (
    ChemFlagError,
    InitialStateError,
    ModelError,
    ModelIO,
    RateConstantError,
    VolumeError,
)
from cayenne.simulation import Simulation


class TestModelIO:
    def test_correct_model_str(self, setup_00001_correct, setup_00001):
        modelio = ModelIO(setup_00001_correct, "ModelString")
        (_, _, react_stoic, prod_stoic, init_state, k_det, *_) = modelio.args
        V_r, V_p, X0, k, *_ = setup_00001
        assert (react_stoic == V_r).all()
        assert (prod_stoic == V_p).all()
        assert (X0 == init_state).all()
        assert (k == k_det).all()

    def test_correct_model_file(self, setup_00001):
        modelio = ModelIO("tests/models/00001.txt", "ModelFile")
        (_, _, react_stoic, prod_stoic, init_state, k_det, *_) = modelio.args
        V_r, V_p, X0, k, *_ = setup_00001
        assert (react_stoic == V_r).all()
        assert (prod_stoic == V_p).all()
        assert (X0 == init_state).all()
        assert (k == k_det).all()

    def test_nochemflag(self, setup_00001_nochemflag):
        with pytest.raises(ChemFlagError):
            ModelIO(setup_00001_nochemflag, "ModelString")

    def test_norate(self, setup_00001_norate):
        with pytest.raises(RateConstantError):
            ModelIO(setup_00001_norate, "ModelString")

    def test_noratevalue(self, setup_00001_noratevalue):
        with pytest.raises(RateConstantError):
            ModelIO(setup_00001_noratevalue, "ModelString")

    def test_rateequation(self, setup_00001_rateequation):
        with pytest.raises(RateConstantError):
            ModelIO(setup_00001_rateequation, "ModelString")

    def test_nospeciesvalue(self, setup_00001_nospeciesvalue):
        with pytest.raises(InitialStateError):
            ModelIO(setup_00001_nospeciesvalue, "ModelString")

    def test_incompletespeciesvalue(self, setup_00001_incompletespeciesvalue):
        with pytest.raises(InitialStateError):
            ModelIO(setup_00001_incompletespeciesvalue, "ModelString")

    def test_nocompartment(self, setup_00001_nocompartment):
        with pytest.raises(VolumeError):
            ModelIO(setup_00001_nocompartment, "ModelString")

    def test_translate_sbml(self):
        sb_string = ModelIO.translate_sbml("tests/models/00001.xml")
        assert isinstance(sb_string, str)



"""
    The class that handles model IO
"""

import numpy as np

import antimony as sb


class ModelError(Exception):
    pass


class InitialStateError(ModelError):
    pass


class RateConstantError(ModelError):
    pass


class ChemFlagError(ModelError):
    pass


class VolumeError(ModelError):
    pass


class ModelIO:
    """
        Class for loading and parsing models

        Parameters
        ---------
        model_contents : str
            Either the model string or the file path
        content_type : str, {"ModelString", "ModelFile"}
            The type of the model

        Attributes
        ----------
        react_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the reactants.
            Reactions are columns and species are rows.
        prod_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the products.
            Reactions are columns and species are rows.
        init_state: (ns,) ndarray
            A 1D array representing the initial state of the system.
        k_det: (nr,) ndarray
            A 1D array representing the deterministic rate constants of the
            system.
        volume: float, optional
            The volume of the reactor vessel which is important for second
            and higher order reactions. Defaults to 1 arbitrary units.
        chem_flag: bool, optional
            If True, divide by Na (Avogadro's constant) while calculating
            stochastic rate constants. Defaults to ``False``.
    """

    def __init__(self, model_contents: str, content_type: str) -> None:
        if content_type == "ModelString":
            er_code = sb.loadAntimonyString(model_contents)
        elif content_type == "ModelFile":
            er_code = sb.loadAntimonyFile(model_contents)
        else:
            raise KeyError(f"Unsupported content_type: {content_type}")
        self.er_code = er_code
        if self.er_code == -1:
            raise ModelError("Error while parsing model")
        self.sb_module = sb.getMainModuleName()
        self._parse_model()

    @staticmethod
    def _create_stoic_mat(ns, nr, name_list, stoic_tuple, species_names):
        """ Function to create the stoichiometric matrix """
        stoic_mat = np.zeros([ns, nr], dtype=np.int)
        for index, (names, stoics) in enumerate(zip(name_list, stoic_tuple)):
            for a_name, a_stoic in zip(names, stoics):
                species_index = species_names.index(a_name)
                stoic_mat[species_index, index] += int(a_stoic)
        return stoic_mat

    def _parse_model(self):
        """ Parse model contents """
        react_stoic_tuple = sb.getReactantStoichiometries(self.sb_module)
        react_names = sb.getReactantNames(self.sb_module)
        prod_names = sb.getProductNames(self.sb_module)
        prod_stoic_tuple = sb.getProductStoichiometries(self.sb_module)
        # 0:all, 1:speciescounts, 2:rateconstants, 6:rxnrateequations, 9:compartmentvols
        species_names = sb.getSymbolNamesOfType(self.sb_module, 1)
        self.species_names = species_names
        rxn_names = sb.getSymbolNamesOfType(self.sb_module, 6)
        self.rxn_names = rxn_names
        ns = len(species_names)
        nr = sb.getNumReactions(self.sb_module)

        # Stochastic matrices
        self.react_stoic = self._create_stoic_mat(
            ns, nr, react_names, react_stoic_tuple, species_names
        )
        self.prod_stoic = self._create_stoic_mat(
            ns, nr, prod_names, prod_stoic_tuple, species_names
        )

        # Initial states
        init_state_values = sb.getSymbolInitialAssignmentsOfType(self.sb_module, 1)
        if "" in init_state_values:
            raise InitialStateError("Missing initial value for one of the species.")
        self.init_state = np.array(init_state_values, dtype=np.int64)

        # Rate constants
        rxn_rateeqns = sb.getSymbolEquationsOfType(self.sb_module, 6)
        rxn_rate_names = list(sb.getSymbolNamesOfType(self.sb_module, 2))
        rxn_rate_values = sb.getSymbolInitialAssignmentsOfType(self.sb_module, 2)
        rxn_rate_dict = dict(zip(rxn_rate_names, rxn_rate_values))

        # Chem flag
        try:
            self.chem_flag = True if rxn_rate_dict["chem_flag"] == "true" else False
        except KeyError:
            raise ChemFlagError("The chem flag was not specified in the model.")

        # Check rate constant specifications
        del rxn_rate_dict["chem_flag"]
        rxn_rate_names.remove("chem_flag")
        print(rxn_rateeqns, rxn_rate_names)
        if "" in rxn_rateeqns:
            raise RateConstantError("Missing rate constant for one of the reactions.")
        for rxn_rateeqn in rxn_rateeqns:
            if rxn_rateeqn not in rxn_rate_names:
                raise RateConstantError(
                    f"{rxn_rateeqn} doesn't match any rate constant."
                )

        # kdet
        try:
            self.k_det = np.array(
                [rxn_rate_dict[rre] for rre in rxn_rateeqns], dtype=float
            )
        except KeyError:
            raise RateConstantError(
                "You are missing a numerical value for one of the rate constants."
            )

        # Volume
        try:
            self.volume = int(
                sb.getSymbolInitialAssignmentsOfType(self.sb_module, 9)[0]
            )
        except IndexError:
            raise VolumeError("Missing compartment information")

    @property
    def args(self):
        """ Returns the attributes of the ModelIO class """
        return (
            self.species_names,
            self.rxn_names,
            self.react_stoic,
            self.prod_stoic,
            self.init_state,
            self.k_det,
            self.chem_flag,
            self.volume,
        )

    @classmethod
    def translate_sbml(cls, sbml_file: str):
        """
            Translate SBML file to Antimony model specification.
            cayenne's model specification is loosely based on Antimony's model
            specification.
        """
        er_code = sb.loadSBMLFile(sbml_file)
        if er_code == -1:
            raise ModelError("Error while parsing model")
        sb_module = sb.getMainModuleName()
        sb_string = sb.getAntimonyString(sb_module)
        return sb_string



"""
    Implementation of the
    `tau leaping algorithm <https://en.wikipedia.org/wiki/Tau-leaping>`_.
    This is an approximate method that needs to be tuned to the system at hand
    (by modifying the time step or the ``tau`` parameter).
    A default ``tau=0.1`` is assumed by
    ``cayenne``. This algorithm is approximate and faster than the Direct
    algorithm, but it must be used with caution. Smaller time steps make the
    simulation more accurate, but increase the code run time. Larger time steps
    makes the simulations less accurate and speeds up code run time.
"""

cimport cython
cimport numpy as np
import numpy as np
import random
from ..utils import get_kstoc, TINY


@cython.boundscheck(False)
@cython.wraparound(False)
def tau_leaping(
    react_stoic: np.ndarray,
    prod_stoic: np.ndarray,
    init_state: np.ndarray,
    k_det: np.ndarray,
    tau: float,
    max_t: float,
    volume: float,
    seed: int,
    chem_flag: bool,
):
    """
        Runs the Tau Leaping Simulation Algorithm.
        Exits if negative population encountered.

        Parameters
        ----------
        react_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the reactants.
            Reactions are columns and species are rows.
        prod_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the products.
            Reactions are columns and species are rows.
        init_state: (ns,) ndarray
            A 1D array representing the initial state of the system.
        k_det: (nr,) ndarray
            A 1D array representing the deterministic rate constants of the
            system.
        tau: float
            The constant time step used to tau leaping.
        max_t: float
            The maximum simulation time to run the simulation for.
        volume: float
            The volume of the reactor vessel which is important for second
            and higher order reactions. Defaults to 1 arbitrary units.
        seed: int
            The seed for the numpy random generator used for the current run
            of the algorithm.
        chem_flag: bool
            If True, divide by Na (Avogadro's constant) while calculating
            stochastic rate constants. Defaults to ``False``.

        Returns
        -------
        t: ndarray
            Numpy array of the times.
        x: ndarray
            Numpy array of the states of the system at times in in `t`.
        status: int
            Indicates the status of the simulation at exit.

            1: Succesful completion, terminated when `max_iter` iterations reached.

            2: Succesful completion, terminated when `max_t` crossed.

            3: Succesful completion, terminated when all species went extinct.

            -1: Failure, order greater than 3 detected.

            -2: Failure, propensity zero without extinction.

            -3: Negative species count encountered.
    """
    cdef:
        int ite = 1, max_iter=0, ind1=0, ind2=0, xtsum=0  # Iteration counter
        double t_curr = 0.0, prop_sum=0.0
        Py_ssize_t ns=react_stoic.shape[0], nr=react_stoic.shape[1]
    v = prod_stoic - react_stoic  # ns x nr
    xt = init_state.copy().astype(np.int64)  # Number of species at time t_curr
    max_iter = np.int(max_t / tau) + 1
    x = np.zeros((max_iter, ns), dtype=np.int64)
    t = np.zeros((max_iter))
    x[0, :] = init_state.copy()
    n_events = np.zeros((nr,), dtype=np.int64)
    np.random.seed(seed)  # Set the seed
    # Determine kstoc from kdet and the highest order or reactions
    prop = get_kstoc(react_stoic, k_det, volume, chem_flag)  # Vector of propensities
    kstoc = prop.copy()  # Stochastic rate constants

    cdef double [:] kstoc_view = kstoc
    cdef double [:] prop_view = prop
    cdef long long [:] xt_view = xt
    cdef long long [:] n_events_view = n_events
    cdef long [:, :] v_view = v
    cdef long [:, :] react_stoic_view = react_stoic
    cdef long long [:, :] x_view = x

    for ind in range(nr):
        prop_sum += prop_view[ind]
    if prop_sum < TINY:
        for ind in range(ns):
            xtsum += xt_view[ind]
        if xtsum > TINY:
            status = -2
            return t[:ite], x[:ite, :], status

    while ite < max_iter:
        # Calculate propensities
        prop_sum = 0
        for ind1 in range(nr):
            for ind2 in range(ns):
                # prop = kstoc * product of (number raised to order)
                if react_stoic_view[ind2, ind1]:
                    if react_stoic_view[ind2, ind1] == 1:
                        prop_view[ind1] *= x_view[ite - 1, ind2]
                    elif react_stoic_view[ind2, ind1] == 2:
                        prop_view[ind1] *= x_view[ite - 1, ind2] * (x_view[ite - 1, ind2] - 1) / 2
                    elif react_stoic_view[ind2, ind1] == 3:
                        prop_view[ind1] *= x_view[ite - 1, ind2] * (x_view[ite - 1, ind2] - 1) * (x_view[ite - 1, ind2] - 2) / 6
            prop_sum += prop_view[ind1]
            n_events_view[ind1] = np.random.poisson(prop_view[ind1] * tau)  # 1 x nr
        if prop_sum == 0:
            status = 3
            return t[:ite], x[:ite,], status
        for ind1 in range(ns):
            xt_view[ind1] = x_view[ite-1, ind1]
            for ind2 in range(nr):
                xt_view[ind1] += n_events_view[ind2] * v_view[ind1, ind2]
        for ind1 in range(ns):
            if xt_view[ind1] < 0:
                return t[:ite], x[:ite, :], -3
            else:
                x_view[ite, ind1] = xt_view[ind1]
        for ind1 in range(nr):
            prop_view[ind1] = kstoc_view[ind1]
        t_curr += tau
        t[ite] = t_curr
        ite += 1
    status = 2
    return t[:ite], x[:ite, :], status



"""
    Tests for direct function
"""

import numpy as np
import pytest

from cayenne.simulation import Simulation


@pytest.mark.parametrize("algorithm", ["direct", "tau_leaping", "tau_adaptive"])
@pytest.mark.usefixtures("setup_basic", "setup_large")
class TestSanitizeAlg:
    """
        Sanity checks on Simulation class where simulations are attempted.
    """

    def test_null(self, algorithm, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        k = np.array([0.0, 0.0])
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm=algorithm)
        assert sim.results.status_list[0] == -2

    def test_status_3(self, algorithm, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        V_p = np.array([[0, 0], [0, 0], [0, 1]])
        X0 = np.array([10, 0, 0], dtype=np.int64)
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm=algorithm, max_t=10, max_iter=100, chem_flag=True)
        assert sim.results.status_list[0] == 3

    def test_status_2(self, algorithm, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        X0 = np.array([10, 0, 0], dtype=np.int64)
        sim = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim.simulate(algorithm=algorithm, max_t=1, max_iter=100, chem_flag=True)
        assert sim.results.status_list[0] == 2

    def test_reproduce(self, algorithm, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        sim1 = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim2 = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim1.simulate(algorithm=algorithm)
        sim2.simulate(algorithm=algorithm)
        assert all(
            (i == j).all() for i, j in zip(sim1.results.t_list, sim2.results.t_list)
        )
        assert all(
            (i == j).all() for i, j in zip(sim1.results.x_list, sim2.results.x_list)
        )
        assert all(
            i == j for i, j in zip(sim1.results.status_list, sim2.results.status_list)
        )

    def test_reproduce_fail(self, algorithm, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        sim1 = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim2 = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim1.simulate(algorithm=algorithm)
        sim2.simulate(algorithm=algorithm, seed=1)
        for i, j in zip(sim1.results.t_list, sim2.results.t_list):
            if i.shape[0] == j.shape[0]:
                assert not (
                    all(
                        (i == j).all()
                        for i, j in zip(sim1.results.t_list, sim2.results.t_list)
                    )
                )

    def test_incorrect_seed(self, algorithm, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        sim1 = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        with pytest.raises(TypeError):
            sim1.simulate(algorithm=algorithm, n_rep=2, seed=[1])
        with pytest.raises(TypeError):
            sim1.simulate(algorithm=algorithm, n_rep=2, seed=[1, 2, 3])

    def test_maxiter_type(self, algorithm, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        sim1 = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        with pytest.raises(TypeError):
            sim1.simulate(algorithm=algorithm, max_iter=100.0)

    def test_monotonic(self, algorithm, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        sim1 = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim1.simulate(algorithm=algorithm)
        results = sim1.results
        for t_array in results.t_list:
            assert (np.diff(t_array) > 0).all()

    def test_long(self, algorithm, setup_long):
        species_names, rxn_names, V_r, V_p, X0, k = setup_long
        sim1 = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        sim1.simulate(
            algorithm=algorithm, max_t=1e5, max_iter=int(1e8), chem_flag=False
        )
        X_output = np.array([0, 0, X0.sum()])
        assert (sim1.results.final[-1] == X_output).all()


@pytest.mark.usefixtures("setup_basic", "setup_large")
class TestSanitize:
    """
        Sanity checks on Simulation class where instance creation fails.
    """

    def test_too_high_order(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        V_r = np.array([[2, 0], [2, 1], [0, 0]])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)

    def test_neg_k(self, setup_large):
        species_names, rxn_names, V_r, V_p, X0, k = setup_large
        k = np.array([1, 1, -1, 1, -1])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)

    def test_k_shape(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        k = np.array([[1], [0], [0]])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)

    def test_Vp_Vr_shape(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        V_p = np.array([[0], [1], [0]])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)

    def test_kdet_Vr_shape(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        k = np.array([1, 1, 1])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)

    def test_X0_Vr_shape(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        X0 = np.array([100, 0, 0, 0])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        X0 = np.array([100, 0])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)

    def test_X0_2d(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        X0 = np.array([[100, 0, 0]])
        print(X0.shape)
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)
        X0 = np.array([[100], [0], [0]])
        print(X0.shape)
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)

    def test_Vp_neg(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        V_p = np.array([[0, 0], [-1, 0], [0, 1]])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)

    def test_Vr_neg(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        V_r = np.array([[-1, 0], [0, 1], [0, 0]])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)

    def test_X0_neg(self, setup_basic):
        species_names, rxn_names, V_r, V_p, X0, k = setup_basic
        X0 = np.array([-10, 0, 0])
        with pytest.raises(ValueError):
            Simulation(species_names, rxn_names, V_r, V_p, X0, k)


# def test_bifurcation(setup_bifurcation):
#     V_r, V_p, X0, k = setup_bifurcation
#     count_excitation = 0
#     n_runs = 10
#     deviation_tolerance = 0.05
#     sim1 = Simulation(species_names, rxn_names, V_r, V_p, X0, k)
#     sim1.simulate(
#         algorithm="tau_adaptive", max_t=150, max_iter=1000, chem_flag=True, n_rep=n_runs
#     )
#     for ind in range(n_runs):
#         x = sim1.results.x_list[ind]
#         xt = x[-1, :]
#         assert np.all(xt - np.array([0, 11, 0, 0]) == 0) or np.all(
#             xt - np.array([0, 0, 1, 10]) == 0
#         )
#         if np.all(xt - np.array([0, 11, 0, 0]) == 0):
#             count_excitation += 1
#     assert np.abs(count_excitation / n_runs - 0.5) < deviation_tolerance


class TestHOR:
    """
        Test the HOR method for various combinations of reactants.
    """

    @staticmethod
    def create_sim_inst(react_stoic):
        prod_stoic = np.random.randint(0, 5, react_stoic.shape)
        k = np.random.rand(react_stoic.shape[1])
        X = np.random.randint(0, 5, react_stoic.shape[0], dtype=np.int64)
        species_names = ["X"] * react_stoic.shape[0]
        rxn_names = ["R"] * react_stoic.shape[1]
        sim = Simulation(species_names, rxn_names, react_stoic, prod_stoic, X, k)
        return sim

    def test_hor1(self):
        # A ->
        react_stoic = np.array([[1, 0]])
        sim = self.create_sim_inst(react_stoic)
        assert sim.HOR == 1

    def test_hor2(self):
        # A ->
        # A ->
        # A ->
        react_stoic = np.array([[1, 1, 1], [0, 0, 0]])
        sim = self.create_sim_inst(react_stoic)
        assert np.all(sim.HOR == [1, 0])

    def test_hor3(self):
        # A + B ->
        # A ->
        react_stoic = np.array([[1, 1], [1, 0]])
        sim = self.create_sim_inst(react_stoic)
        assert np.all(sim.HOR == [2, 2])

    def test_hor4(self):
        # # A + B ->
        # # C ->
        react_stoic = np.array([[1, 0], [1, 0], [0, 1]])
        sim = self.create_sim_inst(react_stoic)
        assert np.all(sim.HOR == [2, 2, 1])

    def test_hor5(self):
        # A + A ->
        # A ->
        # B ->
        react_stoic = np.array([[2, 1, 0], [0, 0, 1]])
        sim = self.create_sim_inst(react_stoic)
        assert np.all(sim.HOR == [-2, 1])

    def test_hor6(self):
        # # A + A + B ->
        # # A + B ->
        # # B ->
        react_stoic = np.array([[2, 1, 0], [1, 1, 1]])
        sim = self.create_sim_inst(react_stoic)
        assert np.all(sim.HOR == [-32, 3])

    def test_hor7(self):
        # A + A + B ->
        # A + B + B ->
        react_stoic = np.array([[2, 1], [1, 2]])
        sim = self.create_sim_inst(react_stoic)
        assert np.all(sim.HOR == [-32, -32])

    def test_hor8(self):
        # # A + A + A ->
        # # A + A + B ->
        # # A + B ->
        react_stoic = np.array([[3, 2, 1], [0, 1, 1]])
        sim = self.create_sim_inst(react_stoic)
        assert np.all(sim.HOR == [-3, 3])

    def test_hor9(self):
        # A + A + A ->
        # B + B + B ->
        # A + A + B ->
        # A + B + B ->
        # A + B + C ->
        react_stoic = np.array([[3, 0, 2, 1, 1], [0, 3, 1, 2, 1], [0, 0, 0, 0, 1]])
        sim = self.create_sim_inst(react_stoic)
        assert np.all(sim.HOR == [-3, -3, 3])

    def test_hor10(self):
        # A + B ->
        # A + A + A ->
        # B + B ->
        # A + C ->
        # C + D + E ->
        react_stoic = np.array(
            [
                [1, 3, 0, 1, 0],
                [1, 0, 2, 0, 0],
                [0, 0, 0, 1, 1],
                [0, 0, 0, 0, 1],
                [0, 0, 0, 0, 1],
            ]
        )
        sim = self.create_sim_inst(react_stoic)
        assert np.all(sim.HOR == [-3, -2, 3, 3, 3])


class TestLoadModel:
    def test_correct_model_str(self, setup_00001_correct):
        sim = Simulation.load_model(setup_00001_correct, "ModelString")
        sim.simulate()
        assert sim.results.status_list == [2]
        sim.simulate(max_t=20.0, max_iter=5)
        assert sim.results.status_list == [1]

    def test_correct_model_file(self):
        sim = Simulation.load_model("tests/models/00001.txt", "ModelFile")
        sim.simulate()
        assert sim.results.status_list == [2]



#!/usr/bin/env python
"""The setup script."""

from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Build import cythonize
import numpy as np

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = ["numpy", "Cython", "matplotlib"]

setup_requirements = ["Cython"]

test_requirements = ["pytest", "pytest-runner", "pytest-benchmark"]

ext_modules = [
    Extension(
        "cayenne.algorithms.direct",
        ["cayenne/algorithms/direct.pyx"],
        include_dirs=[np.get_include()],
    ),
    Extension(
        "cayenne.algorithms.tau_leaping",
        ["cayenne/algorithms/tau_leaping.pyx"],
        include_dirs=[np.get_include()],
    ),
    Extension(
        "cayenne.algorithms.tau_adaptive",
        ["cayenne/algorithms/tau_adaptive.pyx"],
        include_dirs=[np.get_include()],
    ),
    Extension("cayenne.utils", ["cayenne/utils.pyx"], include_dirs=[np.get_include()]),
]

setup(
    author="Dileep Kishore, Srikiran Chandrasekaran",
    author_email="k.dileep1994@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Python package for stochastic simulations",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + "\n\n",
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="cayenne stochastic gillepsie simulation",
    name="cayenne",
    packages=find_packages(exclude=["tests"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/Heuro-labs/cayenne",
    version="1.0.1",
    zip_safe=False,
    ext_modules=cythonize(
        ext_modules,
        annotate=True,
        compiler_directives={"binding": True, "linetrace": False, "profile": False},
    ),
)



"""
    Tests for kstoc function
"""

import numpy as np
import pytest

from cayenne.utils import Na, get_kstoc
from cayenne.simulation import Simulation


@pytest.mark.usefixtures("setup_system")
class TestKstoc:
    """
        Test the output of `get_kstoc`.
    """

    def test_100(self, setup_system):
        """ A -> ..."""
        k_det, volume = setup_system
        V_r = np.array([[1], [0], [0]])
        k_stoc = get_kstoc(V_r, k_det, volume, chem_flag=True)
        assert k_stoc == k_det

    def test_110(self, setup_system):
        """ A + B -> ..."""
        k_det, volume = setup_system
        V_r = np.array([[1], [1], [0]])
        k_stoc = get_kstoc(V_r, k_det, volume, chem_flag=True)
        assert np.isclose(k_stoc, k_det / (Na * volume))

    def test_111(self, setup_system):
        """ A + B + C -> ..."""
        k_det, volume = setup_system
        V_r = np.array([[1], [1], [1]])
        k_stoc = get_kstoc(V_r, k_det, volume, chem_flag=True)
        assert np.isclose(k_stoc, k_det / (Na * volume) ** 2)

    def test_200(self, setup_system):
        """ A + A -> ..."""
        k_det, volume = setup_system
        V_r = np.array([[2], [0], [0]])
        k_stoc = get_kstoc(V_r, k_det, volume, chem_flag=True)
        assert np.isclose(k_stoc, k_det * 2 / (Na * volume))

    def test_210(self, setup_system):
        """ A + A + B -> ..."""
        k_det, volume = setup_system
        V_r = np.array([[2], [1], [0]])
        k_stoc = get_kstoc(V_r, k_det, volume, chem_flag=True)
        assert np.isclose(k_stoc, k_det * 2 / (Na * volume) ** 2)

    def test_300(self, setup_system):
        """ A + A + A -> ..."""
        k_det, volume = setup_system
        V_r = np.array([[3], [0], [0]])
        k_stoc = get_kstoc(V_r, k_det, volume, chem_flag=True)
        assert np.isclose(k_stoc, k_det * 6 / (Na * volume) ** 2)

    def test_chemflag(self, setup_system):
        """ Chemical and non-chemical system """
        k_det, volume = setup_system
        V_r = np.array([[3], [0], [0]])
        k_stoc = get_kstoc(V_r, k_det, volume, chem_flag=True)
        assert np.isclose(k_stoc, k_det * 6 / (Na * volume) ** 2)
        k_det, volume = setup_system
        V_r = np.array([[3], [0], [0]])
        k_stoc = get_kstoc(V_r, k_det, volume, chem_flag=False)
        assert np.isclose(k_stoc, k_det * 6 / (1 * volume) ** 2)



"""
    Tests for plotting function
"""


from cayenne.simulation import Simulation


def test_plotting(setup_basic):
    """ Test if plot generates without errors """
    species_name, rxn_names, V_r, V_p, X0, k = setup_basic
    n_runs = 10
    sim1 = Simulation(species_name, rxn_names, V_r, V_p, X0, k)
    sim1.simulate(
        algorithm="direct", max_t=150, max_iter=1000, chem_flag=True, n_rep=n_runs
    )
    sim1.plot(species_names=["A"])
    sim1.plot(species_names=["A", "B"], new_names=["Species A", "Secies B"])
    sim1.plot()



"""
    The main class for running stochastic simulation
"""

import multiprocessing as mp
from functools import partial
from typing import List, Optional
from warnings import warn

import matplotlib.lines as mlines
import matplotlib.pyplot as plt
import numpy as np

from .algorithms.direct import direct
from .algorithms.tau_adaptive import tau_adaptive
from .algorithms.tau_leaping import tau_leaping
from .model_io import ModelIO
from .results import Results


def wrapper(x, func):
    return func(*x)


class Simulation:
    """
        A main class for running simulations.

        Parameters
        ----------
        species_names : List[str]
            List of species names
        rxn_names : List[str]
            List of reaction names
        react_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the reactants.
            Reactions are columns and species are rows.
        prod_stoic: (ns, nr) ndarray
            A 2D array of the stoichiometric coefficients of the products.
            Reactions are columns and species are rows.
        init_state: (ns,) ndarray
            A 1D array representing the initial state of the system.
        k_det: (nr,) ndarray
            A 1D array representing the deterministic rate constants of the
            system.
        volume: float, optional
            The volume of the reactor vessel which is important for second
            and higher order reactions. Defaults to 1 arbitrary units.
        chem_flag: bool, optional
            If True, divide by Na (Avogadro's constant) while calculating
            stochastic rate constants. Defaults to ``False``.

        Attributes
        ----------
        results: Results
            The results instance

        Raises
        ------
        ValueError
            If supplied with order > 3.

        Examples
        --------
        >>> V_r = np.array([[1,0],[0,1],[0,0]])
        >>> V_p = np.array([[0,0],[1,0],[0,1]])
        >>> X0 = np.array([10,0,0])
        >>> k = np.array([1,1])
        >>> sim = Simulation(V_r, V_p, X0, k)

        Notes
        -----
        Stochastic reaction rates depend on the size of the system for second
        and third order reactions. By this, we mean the volume of the system in
        which the reactants are contained. Intuitively, this makes sense
        considering that collisions between two or more molecules becomes
        rarer as the size of the system increases. A detailed mathematical
        treatment of this idea can be found in [3]_ .

        In practice, this means that ``volume`` and ``chem_flag`` need to be
        supplied for second and third order reactions. ``volume``
        represents the size of the system containing the reactants.

        In chemical systems ``chem_flag`` should generally be set to ``True``
        as ``k_det`` is specified in units of molarity or M or mol/L.
        For example, a second order rate constant could be = 0.15 mol / (L s).
        Then Avogadro's constant (:math:`N_a`) is used for normalization while
        computing ``k_stoc`` (:math:`c_\mu` in [3]_ ) from ``k_det``.

        In biological systems, ``chem_flag`` should be generally be set to
        ``False`` as ``k_det`` is specified in units of copies/L or CFU/L.
        For example, a second order rate constant could be = 0.15 CFU / (L s).

        References
        ----------
        .. [3] Gillespie, D.T., 1976.
            A general method for numerically simulating the stochastic time
            evolution of coupled chemical reactions. J. Comput. Phys. 22, 403–434.
            doi:10.1016/0021-9991(76)90041-3.

    """

    _results: Optional[Results] = None

    def __init__(
        self,
        species_names: List[str],
        rxn_names: List[str],
        react_stoic: np.ndarray,
        prod_stoic: np.ndarray,
        init_state: np.ndarray,
        k_det: np.ndarray,
        chem_flag: bool = False,
        volume: float = 1.0,
    ) -> None:
        self.species_names = species_names
        self.rxn_names = rxn_names
        self._react_stoic = react_stoic
        self._prod_stoic = prod_stoic
        self._init_state = init_state
        self._k_det = k_det
        self._chem_flag = chem_flag
        self._ns = self._react_stoic.shape[0]
        self._nr = self._react_stoic.shape[1]
        self._volume = volume
        self._orders = np.sum(
            self._react_stoic, axis=0
        )  # Order of rxn = number of reactants
        self._check_consistency()

    def _check_consistency(self):
        if len(self.species_names) != self._ns:
            raise ValueError(
                "Species names must match the stoichiometric matrix shapes."
            )
        if len(self.rxn_names) != self._nr:
            raise ValueError(
                "Reaction names must match the stoichiometric matrix shapes."
            )
        if (self._ns != self._prod_stoic.shape[0]) or (
            self._nr != self._prod_stoic.shape[1]
        ):
            raise ValueError("react_stoic and prod_stoic should be the same shape.")
        if np.any(self._react_stoic < 0):
            raise ValueError("react_stoic cannot have negative elements.")
        if np.any(self._prod_stoic < 0):
            raise ValueError("V_p cannot have negative elements.")
        if np.any(self._init_state < 0):
            raise ValueError("Initial numbers in X0 can't be negative.")
        if np.any(self._k_det < 0):
            raise ValueError("Rate constant(s) can't be negative.")
        if self._k_det.shape[0] != self._nr:
            raise ValueError("Number of rate constants must equal number of reactions")
        if len(self._k_det.shape) != 1:
            raise ValueError("k_det must be a 1-D array")
        if self._chem_flag not in (True, False):
            raise ValueError("chem_flag must be a boolean True or False.")
        if np.max(self._orders) > 3:
            raise ValueError("Order greater than 3 not suppported.")
        if self._ns != self._init_state.shape[0]:
            raise ValueError(
                "X0 must have be of length = num. of species (or rows of V_r)"
            )
        if len(self._init_state.shape) != 1:
            raise ValueError("X0 must be a 1-D array")

    @property
    def results(self) -> Optional[Results]:
        """
            The ``Results`` instance of the simulation

            Returns
            -------
            Optional[Results]
        """
        if self._results is None:
            warn(
                "Run `Simulation.simulate` before requesting the results object",
                Warning,
            )
            return self._results
        else:
            return self._results

    @classmethod
    def load_model(cls, contents: str, contents_type: str) -> "Simulation":
        """
            Load model contents into a Simulation object

            Parameters
            ----------
            model_contents : str
                Either the model string or the file path
            content_type : str, {"ModelString", "ModelFile"}
                The type of the model

            Returns
            -------
            sim : Simulation
                An instance of the Simulation class.
         """
        modelio = ModelIO(contents, contents_type)
        (
            species_names,
            rxn_names,
            react_stoic,
            prod_stoic,
            init_state,
            k_det,
            chem_flag,
            volume,
        ) = modelio.args
        return cls(
            species_names,
            rxn_names,
            react_stoic,
            prod_stoic,
            init_state,
            k_det,
            chem_flag,
            volume,
        )

    def simulate(
        self,
        max_t: float = 10.0,
        max_iter: int = 1000,
        seed: int = 0,
        n_rep: int = 1,
        n_procs: Optional[int] = 1,
        algorithm: str = "direct",
        debug: bool = False,
        **kwargs,
    ) -> None:
        """
            Run the simulation

            Parameters
            ----------
            max_t: float, optional
                The end time of the simulation.
                The default is 10.0.
            max_iter: int, optional
                The maximum number of iterations of the simulation loop.
                The default is 1000 iterations.
            seed: int, optional
                The seed used to generate simulation seeds.
                The default value is 0.
            n_rep: int, optional
                The number of repetitions of the simulation required.
                The default value is 1.
            n_procs: int, optional
                The number of cpu cores to use for the simulation.
                Use ``None`` to automatically detect number of cpu cores.
                The default value is 1.
            algorithm: str, optional
                The algorithm to be used to run the simulation.
                The default value is ``"direct"``.

            Notes
            -----
            The status indicates the status of the simulation at exit. Each
            repetition will have a status associated with it, and these are
            accessible through the ``Simulation.results.status_list``.

            status: int
                Indicates the status of the simulation at exit.

                1: Succesful completion, terminated when ``max_iter`` iterations reached.

                2: Succesful completion, terminated when ``max_t`` crossed.

                3: Succesful completion, terminated when all species went extinct.

                -1: Failure, order greater than 3 detected.

                -2: Failure, propensity zero without extinction.
        """
        tlist = []
        xlist = []
        status_list = []

        if not isinstance(seed, int):
            raise TypeError("Seed should be of type int")
        np.random.seed(seed)
        sim_seeds = np.random.randint(low=0, high=1e7, size=n_rep)

        if not isinstance(max_iter, int):
            raise TypeError("max_iter should be of type int")

        algo_args = []
        if algorithm == "direct":
            for index in range(n_rep):
                algo_args.append(
                    (
                        self._react_stoic,
                        self._prod_stoic,
                        self._init_state,
                        self._k_det,
                        max_t,
                        max_iter,
                        self._volume,
                        sim_seeds[index],
                        self._chem_flag,
                    )
                )
                algo = direct
        elif algorithm == "tau_leaping":
            if "tau" in kwargs.keys():
                tau = kwargs["tau"]
            else:
                tau = 0.1
            for index in range(n_rep):
                algo_args.append(
                    (
                        self._react_stoic,
                        self._prod_stoic,
                        self._init_state,
                        self._k_det,
                        tau,
                        max_t,
                        self._volume,
                        sim_seeds[index],
                        self._chem_flag,
                    )
                )
                algo = tau_leaping
        elif algorithm == "tau_adaptive":
            if "epsilon" in kwargs.keys():
                epsilon = kwargs["epsilon"]
            else:
                epsilon = 0.03
            if "nc" in kwargs.keys():
                nc = kwargs["nc"]
            else:
                nc = 10
            hor = self.HOR
            for index in range(n_rep):
                algo_args.append(
                    (
                        self._react_stoic,
                        self._prod_stoic,
                        self._init_state,
                        self._k_det,
                        hor.astype(np.int),
                        nc,
                        epsilon,
                        max_t,
                        max_iter,
                        self._volume,
                        sim_seeds[index],
                        self._chem_flag,
                    )
                )
            algo = tau_adaptive
        else:
            raise ValueError("Requested algorithm not supported")
        algo_func = partial(wrapper, func=algo)
        if debug:
            results = map(algo_func, algo_args)
        else:
            with mp.Pool(processes=n_procs) as pool:
                results = pool.map(algo_func, algo_args)
        for t, X, status in results:
            tlist.append(t)
            xlist.append(X)
            status_list.append(status)
        self._results = Results(
            self.species_names,
            self.rxn_names,
            tlist,
            xlist,
            status_list,
            algorithm,
            sim_seeds,
        )

    @property
    def HOR(self) -> np.ndarray:
        """
            Determine the HOR vector. HOR(i) is the highest order of reaction
            in which species S_i appears as a reactant.

            Returns
            -------
            HOR: np.ndarray
                Highest order of the reaction for the reactive species as
                defined under Eqn. (27) of [1]_. HOR can be 1, 2 or 3
                if the species appears only once in the reactants.
                If HOR is -2, it appears twice in a second order reaction.
                If HOR is -3, it appears thrice in a third order reaction.
                If HOR is -32, it appears twice in a third order reaction.
                The corresponding value of ``g_i`` in Eqn. (27) is handled
                by ``tau_adaptive``.

            References
            ----------
            .. [1] Cao, Y., Gillespie, D.T., Petzold, L.R., 2006.
                Efficient step size selection for the tau-leaping simulation
                method. J. Chem. Phys. 124, 044109. doi:10.1063/1.2159468
        """
        ns = self._react_stoic.shape[0]
        HOR = np.zeros(ns, dtype=np.int32)
        orders = np.sum(self._react_stoic, axis=0)
        for ind in range(ns):
            this_orders = orders[np.where(self._react_stoic[ind, :] > 0)[0]]
            if len(this_orders) == 0:
                HOR[ind] = 0
                continue
            HOR[ind] = np.max(this_orders)
            if HOR[ind] == 1:
                continue
            order_2_indices = np.where(orders == 2)
            this_react_stoic = self._react_stoic[ind, :]
            if order_2_indices[0].size > 0:
                if np.max(this_react_stoic[order_2_indices[0]]) == 2 and HOR[ind] == 2:
                    HOR[ind] = -2  # g_i should be (2 + 1/(x_i-1))
            if np.where(orders == 3):
                if (
                    HOR[ind] == 3
                    and np.max(this_react_stoic[np.where(this_orders == 3)[0]]) == 2
                ):
                    HOR[ind] = -32  # g_i should be (3/2 * (2 + 1/(x_i-1)))
                elif (
                    HOR[ind] == 3
                    and np.max(this_react_stoic[np.where(this_orders == 3)[0]]) == 3
                ):
                    HOR[ind] = -3  # g_i should be(3 + 1/(x_i-1) + 2/(x_i-2))
        return HOR

    def plot(self, species_names: list = None, new_names: list = None):
        """
            Plot the simulation

            Parameters
            ----------
            species_names: list, optional
                The names of the species to be plotted (``list`` of ``str``).
                The default is ``None`` and  plots all species.
            new_names: list, optional
                The names of the species to be plotted.
                The default is ``"xi"`` for species ``i``.

            Returns
            -------
            fig: class 'matplotlib.figure.Figure'
                Figure object of the generated plot.
            ax: class 'matplotlib.axes._subplots.AxesSubplot
                Axis objected of the generated plot.
        """

        if self._results is None:
            raise ValueError("Simulate not run.")
        else:
            if species_names is None:
                species_names = self.species_names
            n_species = len(species_names)
            prop_cycle = plt.rcParams["axes.prop_cycle"]
            colors = prop_cycle.by_key()["color"]
            fig, ax = plt.subplots()
            res = self._results
            legend_handlers = [0] * n_species
            generic_names = [""] * n_species
            for index1 in range(n_species):
                legend_handlers[index1] = mlines.Line2D([], [], color=colors[index1])
                this_species = species_names[index1]
                generic_names[index1] = this_species
                n_reps = len(res)
                for index2 in range(n_reps):
                    ax.step(
                        res.t_list[index2],
                        res.get_species([this_species])[index2],
                        color=colors[index1],
                        where="post",
                    )
            if new_names is None:
                new_names = self.species_names
            fig.legend(legend_handlers, new_names)
            ax.set_xlabel("Time")
            ax.set_ylabel("Species amounts")
            return fig, ax



"""
    Tests for tau leaping with adaptive step size selection function
"""

import numpy as np
from cayenne.algorithms.tau_adaptive import (
    _py_step1,
    _py_step2,
    _py_step2_get_g,
    _py_step5,
)
from cayenne.utils import get_kstoc, HIGH
from .test_simulation import TestHOR


def test_tauadaptive_step1(setup_basic):
    _, _, V_r, V_p, X0, k = setup_basic
    V = V_p - V_r
    kstoc = get_kstoc(V_r, k, 1.0, False)
    X0 = np.array([100, 9, 0], dtype=np.int64)
    prop, crit, not_crit = _py_step1(kstoc, X0, V_r, V, nc=10)
    print(prop, crit, not_crit)
    assert (prop == [X0[0], X0[1]]).all()
    assert (crit != not_crit).all()
    assert (crit == np.array([False, True])).all()

    X0 = np.array([100, 11, 0], dtype=np.int64)
    prop, crit, not_crit = _py_step1(kstoc, X0, V_r, V, nc=10)
    print(prop, crit, not_crit)
    assert (prop == [X0[0], X0[1]]).all()
    assert (crit == np.array([False, False])).all()

    X0 = np.array([100, 0, 0], dtype=np.int64)
    prop, crit, not_crit = _py_step1(kstoc, X0, V_r, V, nc=10)
    print(prop, crit, not_crit)
    assert (prop == [X0[0], X0[1]]).all()
    assert (crit == np.array([False, False])).all()

    X0 = np.array([0, 9, 0], dtype=np.int64)
    prop, crit, not_crit = _py_step1(kstoc, X0, V_r, V, nc=10)
    print(prop, crit, not_crit)
    assert (prop == [X0[0], X0[1]]).all()
    assert (crit == np.array([False, True])).all()

    X0 = np.array([0, 0, 0], dtype=np.int64)
    prop, crit, not_crit = _py_step1(kstoc, X0, V_r, V, nc=10)
    print(prop, crit, not_crit)
    assert (prop == [X0[0], X0[1]]).all()
    assert (crit == np.array([False, False])).all()

    # 0 -> A
    # A -> 0
    V_r = np.array([[0, 1]])
    V_p = np.array([[1, 0]])
    V = V_p - V_r
    X0 = np.array([100], dtype=np.int64)
    prop, crit, not_crit = _py_step1(kstoc, X0, V_r, V, nc=10)
    assert (crit == np.array([False, False])).all()

    X0 = np.array([9], dtype=np.int64)
    prop, crit, not_crit = _py_step1(kstoc, X0, V_r, V, nc=10)
    assert (crit == np.array([False, True])).all()

    X0 = np.array([0], dtype=np.int64)
    prop, crit, not_crit = _py_step1(kstoc, X0, V_r, V, nc=10)
    assert (crit == np.array([False, False])).all()


# def test_len_sim(setup_basic):
#     V_r, V_p, X0, k = setup_basic
#     X0 = X0 = np.array([50, 2, 0])
#     sim = Simulation(V_r, V_p, X0, k)
#     sim.simulate(algorithm="tau_adaptive")
#     x = sim.results.x_list
#     t = sim.results.t_list
#     assert ~np.all(np.array(x[0][0, :]) == np.array(x[0][1, :]))
#     assert ~np.all(np.array(t[0][0]) == np.array(t[0][1]))


def test_tauadaptive_step2(setup_basic):
    _, _, V_r, V_p, X0, k = setup_basic
    V = V_p - V_r
    sim = TestHOR.create_sim_inst(V_r)
    hor = np.int32(sim.HOR)
    assert np.all(hor == [1, 1, 0])
    react_species = np.int32(np.where(np.sum(V_r, axis=1) > 0)[0])

    assert (react_species == np.array([0, 1])).all()
    kstoc = get_kstoc(V_r, k, 1.0, False)
    prop, _, not_crit = _py_step1(kstoc, np.int64(X0), V_r, V, nc=10)
    taup = _py_step2(not_crit, react_species, V, np.int64(X0), hor, prop, epsilon=0.03)
    assert taup == 0.01

    taup = _py_step2(
        np.zeros(not_crit.shape, dtype=np.int32),
        react_species,
        V,
        np.int64(X0),
        hor,
        prop,
        epsilon=0.03,
    )
    assert np.isclose(taup, HIGH)


def test_tauadaptive_step2_get_g():
    assert _py_step2_get_g(1, 10) == 1
    assert _py_step2_get_g(-2, 2) == 3
    assert _py_step2_get_g(-2, 1) == 2
    assert np.isclose(_py_step2_get_g(-2, 10), 2 + (1 / 9))
    assert _py_step2_get_g(-3, 1) == 3.0
    assert np.isclose(_py_step2_get_g(-3, 10), 3 + 1 / 9 + 2 / 8)
    assert _py_step2_get_g(-32, 1) == 3.0
    assert np.isclose(_py_step2_get_g(-32, 10), 3 / 2 * (2 + 1 / 9))


def test_tauadaptive_step5(setup_basic):
    _, _, V_r, V_p, X0, k = setup_basic
    nr = V_r.shape[1]
    V = V_p - V_r
    kstoc = get_kstoc(V_r, k, 1.0, False)
    prop, _, not_crit = _py_step1(kstoc, np.int64(X0), V_r, V, nc=10)
    taup, taupp = 100.0, 10.0
    tau, K = _py_step5(taup, taupp, nr, not_crit, prop, np.int64(X0))
    assert tau == taupp
    assert K[1] == 0
    taup, taupp = 0.01, 1.0
    tau, K = _py_step5(taup, taupp, nr, not_crit, prop, np.int64(X0))
    assert tau == taup
    assert K[1] == 0
